#! /usr/bin/env python

import openturns as ot

ot.TESTPREAMBLE()


def cleanSymmetricTensor(inSymmetricTensor):
    rowDim = inSymmetricTensor.getNbRows()
    colDim = inSymmetricTensor.getNbColumns()
    sheetDim = inSymmetricTensor.getNbSheets()
    for i in range(rowDim):
        for j in range(colDim):
            for k in range(sheetDim):
                inSymmetricTensor[i, j, k] = 1.0e-4 * round(
                    1.0e4 * inSymmetricTensor[i, j, k]
                )
                if abs(inSymmetricTensor[i, j, k]) < 1.0e-6:
                    inSymmetricTensor[i, j, k] = 0.0
    return inSymmetricTensor


# Create a collection of distribution
aCollection = ot.DistributionCollection()

aCollection.add(ot.Uniform(-1.0, 2.0))
aCollection.add(ot.Gamma(2.0, 2.0, 0.0))

dim = aCollection.getSize()

# Create a copula
RCopula = ot.CorrelationMatrix(dim)
for i in range(dim):
    for j in range(i):
        RCopula[i, j] = (i + j + 1.0) / (2.0 * dim)

# Instantiate one distribution object
distribution = ot.JointDistribution(aCollection, ot.NormalCopula(RCopula))
# Test for sampling
size = 10000
sample = distribution.getSample(size)
print("sample first=", repr(sample[0]), " last=", repr(sample[size - 1]))
print("sample mean=", repr(sample.computeMean()))
print("sample covariance=", repr(sample.computeCovariance()))

transform = distribution.getIsoProbabilisticTransformation()
print("isoprobabilistic transformation=", repr(transform))
transformedSample = transform(sample)
print(
    "transformed sample first=",
    repr(transformedSample[0]),
    " last=",
    repr(transformedSample[size - 1]),
)
print("transformed sample mean=", repr(transformedSample.computeMean()))
print("transformed sample covariance=", repr(transformedSample.computeCovariance()))

# Test for evaluation
inverseTransform = distribution.getInverseIsoProbabilisticTransformation()
print("inverse isoprobabilistic transformation=", repr(inverseTransform))
transformedBackSample = inverseTransform(transformedSample)
print(
    "transformed back sample first=",
    repr(transformedBackSample[0]),
    " last=",
    repr(transformedBackSample[size - 1]),
)
print("transformed back sample mean=", repr(transformedBackSample.computeMean()))
print(
    "transformed back sample covariance=",
    repr(transformedBackSample.computeCovariance()),
)
point = ot.Point(dim, 1.0)
print("point=", repr(point))
transformedPoint = transform(point)
print("transform value at point        =", repr(transformedPoint))
print("transform gradient at point     =", repr(transform.gradient(point)))
print(
    "transform gradient at point (FD)=",
    repr(
        ot.CenteredFiniteDifferenceGradient(1.0e-5, transform.getEvaluation()).gradient(
            point
        )
    ),
)
print(
    "transform hessian at point      =",
    repr(cleanSymmetricTensor(transform.hessian(point))),
)
print(
    "transform hessian at point (FD) =",
    repr(
        cleanSymmetricTensor(
            ot.CenteredFiniteDifferenceHessian(
                1.0e-4, transform.getEvaluation()
            ).hessian(point)
        )
    ),
)
print(
    "inverse transform value at transformed point        =",
    repr(inverseTransform(transformedPoint)),
)
print(
    "inverse transform gradient at transformed point     =",
    repr(inverseTransform.gradient(transformedPoint)),
)
print(
    "inverse transform gradient at transformed point (FD)=",
    repr(
        ot.CenteredFiniteDifferenceGradient(
            1.0e-5, inverseTransform.getEvaluation()
        ).gradient(transformedPoint)
    ),
)
print(
    "inverse transform hessian at transformed point      =",
    repr(cleanSymmetricTensor(inverseTransform.hessian(transformedPoint))),
)
print(
    "inverse transform hessian at transformed point (FD) =",
    repr(
        cleanSymmetricTensor(
            ot.CenteredFiniteDifferenceHessian(
                1.0e-4, inverseTransform.getEvaluation()
            ).hessian(transformedPoint)
        )
    ),
)

# Test for parameters
print("parameters gradient at point=", repr(transform.parameterGradient(point)))

# Validation using finite difference
eps = 1e-5
factor = 1.0 / (2.0 * eps)
gradient = ot.Matrix(5, 2)

# dT/dp0
coll = ot.DistributionCollection(dim)
coll[0] = ot.Uniform(-1.0 + eps, 2.0)
coll[1] = aCollection[1]
left = ot.JointDistribution(
    coll, ot.NormalCopula(RCopula)
).getIsoProbabilisticTransformation()
coll[0] = ot.Uniform(-1.0 - eps, 2.0)
right = ot.JointDistribution(
    coll, ot.NormalCopula(RCopula)
).getIsoProbabilisticTransformation()
dTdp = (left(point) - right(point)) * factor
gradient[0, 0] = dTdp[0]
gradient[0, 1] = dTdp[1]
# dT/dp1
coll = ot.DistributionCollection(dim)
coll[0] = ot.Uniform(-1.0, 2.0 + eps)
coll[1] = aCollection[1]
left = ot.JointDistribution(
    coll, ot.NormalCopula(RCopula)
).getIsoProbabilisticTransformation()
coll[0] = ot.Uniform(-1.0, 2.0 - eps)
right = ot.JointDistribution(
    coll, ot.NormalCopula(RCopula)
).getIsoProbabilisticTransformation()
dTdp = (left(point) - right(point)) * factor
gradient[1, 0] = dTdp[0]
gradient[1, 1] = dTdp[1]
# dT/dp2
coll = ot.DistributionCollection(dim)
coll[0] = aCollection[0]
coll[1] = ot.Gamma(2.0 + eps, 2.0, 0.0)
left = ot.JointDistribution(
    coll, ot.NormalCopula(RCopula)
).getIsoProbabilisticTransformation()
coll[1] = ot.Gamma(2.0 - eps, 2.0, 0.0)
right = ot.JointDistribution(
    coll, ot.NormalCopula(RCopula)
).getIsoProbabilisticTransformation()
dTdp = (left(point) - right(point)) * factor
gradient[2, 0] = dTdp[0]
gradient[2, 1] = dTdp[1]
# dT/dp3
coll = ot.DistributionCollection(dim)
coll[0] = aCollection[0]
coll[1] = ot.Gamma(2.0, 2.0 + eps, 0.0)
left = ot.JointDistribution(
    coll, ot.NormalCopula(RCopula)
).getIsoProbabilisticTransformation()
coll[1] = ot.Gamma(2.0, 2.0 - eps, 0.0)
right = ot.JointDistribution(
    coll, ot.NormalCopula(RCopula)
).getIsoProbabilisticTransformation()
dTdp = (left(point) - right(point)) * factor
gradient[3, 0] = dTdp[0]
gradient[3, 1] = dTdp[1]
# dT/dp4
coll = ot.DistributionCollection(dim)
coll[0] = aCollection[0]
coll[1] = ot.Gamma(2.0, 2.0, 0.0 + eps)
left = ot.JointDistribution(
    coll, ot.NormalCopula(RCopula)
).getIsoProbabilisticTransformation()
coll[1] = ot.Gamma(2.0, 2.0, 0.0 - eps)
right = ot.JointDistribution(
    coll, ot.NormalCopula(RCopula)
).getIsoProbabilisticTransformation()
dTdp = (left(point) - right(point)) * factor
gradient[4, 0] = dTdp[0]
gradient[4, 1] = dTdp[1]

print("parameters gradient (FD)    =", repr(gradient))

# Test for parameters
print(
    "(inverse) parameters gradient at point=",
    repr(inverseTransform.parameterGradient(point)),
)

# dT/dp0
coll = ot.DistributionCollection(dim)
coll[0] = ot.Uniform(-1.0 + eps, 2.0)
coll[1] = aCollection[1]
left = ot.JointDistribution(
    coll, ot.NormalCopula(RCopula)
).getInverseIsoProbabilisticTransformation()
coll[0] = ot.Uniform(-1.0 - eps, 2.0)
right = ot.JointDistribution(
    coll, ot.NormalCopula(RCopula)
).getInverseIsoProbabilisticTransformation()
dTdp = (left(point) - right(point)) * factor
gradient[0, 0] = dTdp[0]
gradient[0, 1] = dTdp[1]
# dT/dp1
coll = ot.DistributionCollection(dim)
coll[0] = ot.Uniform(-1.0, 2.0 + eps)
coll[1] = aCollection[1]
left = ot.JointDistribution(
    coll, ot.NormalCopula(RCopula)
).getInverseIsoProbabilisticTransformation()
coll[0] = ot.Uniform(-1.0, 2.0 - eps)
right = ot.JointDistribution(
    coll, ot.NormalCopula(RCopula)
).getInverseIsoProbabilisticTransformation()
dTdp = (left(point) - right(point)) * factor
gradient[1, 0] = dTdp[0]
gradient[1, 1] = dTdp[1]
# dT/dp2
coll = ot.DistributionCollection(dim)
coll[0] = aCollection[0]
coll[1] = ot.Gamma(2.0 + eps, 2.0, 0.0)
left = ot.JointDistribution(
    coll, ot.NormalCopula(RCopula)
).getInverseIsoProbabilisticTransformation()
coll[1] = ot.Gamma(2.0 - eps, 2.0, 0.0)
right = ot.JointDistribution(
    coll, ot.NormalCopula(RCopula)
).getInverseIsoProbabilisticTransformation()
dTdp = (left(point) - right(point)) * factor
gradient[2, 0] = dTdp[0]
gradient[2, 1] = dTdp[1]
# dT/dp3
coll = ot.DistributionCollection(dim)
coll[0] = aCollection[0]
coll[1] = ot.Gamma(2.0, 2.0 + eps, 0.0)
left = ot.JointDistribution(
    coll, ot.NormalCopula(RCopula)
).getInverseIsoProbabilisticTransformation()
coll[1] = ot.Gamma(2.0, 2.0 - eps, 0.0)
right = ot.JointDistribution(
    coll, ot.NormalCopula(RCopula)
).getInverseIsoProbabilisticTransformation()
dTdp = (left(point) - right(point)) * factor
gradient[3, 0] = dTdp[0]
gradient[3, 1] = dTdp[1]
# dT/dp4
coll = ot.DistributionCollection(dim)
coll[0] = aCollection[0]
coll[1] = ot.Gamma(2.0, 2.0, 0.0 + eps)
left = ot.JointDistribution(
    coll, ot.NormalCopula(RCopula)
).getInverseIsoProbabilisticTransformation()
coll[1] = ot.Gamma(2.0, 2.0, 0.0 - eps)
right = ot.JointDistribution(
    coll, ot.NormalCopula(RCopula)
).getInverseIsoProbabilisticTransformation()
dTdp = (left(point) - right(point)) * factor
gradient[4, 0] = dTdp[0]
gradient[4, 1] = dTdp[1]

print("(inverse) parameters gradient (FD)    =", repr(gradient))



#! /usr/bin/env python

import openturns as ot

ot.TESTPREAMBLE()

# Analytical construction */
input = ot.Description(1)
input[0] = "x"
formulas = ot.Description(1)
formulas[0] = "x-cos(2*x)"
analytical = ot.SymbolicFunction(input, formulas)
myAlgo = ot.Secant()
print("myAlgo=", myAlgo)
value = 0.0
print("Solve ", formulas[0], "=%.5f" % value, " for ", input[0])
print("x=%.5f" % myAlgo.solve(analytical, value, 0.0, 3.0))
print("myAlgo=", myAlgo)



"""
Estimate a multivariate integral with IteratedQuadrature
========================================================
"""

# %%
#
# Introduction
# ------------
#
# In this example, we consider a function :math:`f: \Rset^d \mapsto \Rset`
# and we compute the following integral:
#
# .. math::
#    I_f = \int_{a}^{b}\, \int_{l_1(x_0)}^{u_1(x_0)}\, \int_{l_2(x_0, x_1)}^{u_2(x_0,x_1)}\, \int_{l_{d-1}(x_0, \dots, x_{d-2})}^{u_{d-1}(x_0, \dots, x_{d-2})}
#    \, f(x_0, \dots, x_{d-1})\mathrm{d}{x_{d-1}}\dots\mathrm{d}{x_0}
#
# using an iterated quadrature algorithm :class:`~openturns.IteratedQuadrature`, based on the Gauss-Kronrod
# :class:`~openturns.GaussKronrod` 1d quadrature algorithm.

import openturns as ot
import openturns.viewer as otv
import math as m

# %%
# To get better colours for the function iso-lines.
ot.ResourceMap.SetAsString("Contour-DefaultColorMapNorm", "rank")

# %%
# Case 1: The integrand function presents a lot of discontinuities.
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# We consider the function:
#
# .. math::
#    f : \begin{array}{lcl}
#          \Rset^2 & \rightarrow & \Rset \\
#          (x,y) & \rightarrow & 1_{{x^2+y^2 \leq 4}}(x,y)
#          \end{array}
#
# the domain :math:`\set{D} =[-2,2]\times [-2,2]` and the integral:
#
# .. math::
#    I = \int_{\set{D}} f(x,y)\,dxdy = 4\pi
#

# %%
# We first define the integrand :math:`f` and the domain :math:`\set{D}`.


def kernel1(x):
    if x[0] ** 2 + x[1] ** 2 <= 4:
        return [1.0]
    else:
        return [0.0]


integrand = ot.PythonFunction(2, 1, kernel1)
domain = ot.Interval([-2.0, -2.0], [2.0, 2.0])

# %%
# We draw the iso-lines of the integrand function which is
# constant equal to 1 inside the circle with radius equal to 2 and equal to 0.0 outside.
g = integrand.draw([-3.0, -3.0], [3.0, 3.0])
g.setTitle(r"Function $f(x,y) = 1_{\{x^2+y^2 \leq 4\}}(x,y)$")
g.setXTitle("x")
g.setYTitle("y")
view = otv.View(g)

# %%
# We compute the integral using an  :class:`~openturns.IteratedQuadrature`
# algorithm with
# the :class:`~openturns.GaussKronrodRule`  *G11K23*, which means that we use
# on each subinterval:
#
# - 11 points for the Gauss approximations,
# - 23 points for the Kronrod approximation including the 11 previous ones.
#
# We limit the number of subintervals and we  define a maximum error between both approximations:
# when the difference between both approximations is lower, a new subdivision is performed.
maximumSubIntervals = 32
maximumError = 1e-4
GKRule = ot.GaussKronrodRule(ot.GaussKronrodRule.G1K3)
integ_algo = ot.GaussKronrod(maximumSubIntervals, maximumError, GKRule)
integration_algo = ot.IteratedQuadrature(integ_algo)

# %%
# To get the nodes used to approximate the integral, we need to use a :class:`~openturns.MemoizeFunction`
# that stores all the calls to the function.
integrand_memoized = ot.MemoizeFunction(integrand)
I_value = integration_algo.integrate(integrand_memoized, domain)
print("I = ", I_value[0])
print("Exact value = ", 4 * m.pi)
nodes = integrand_memoized.getInputHistory()
print("Nodes : ", nodes)

# %%
# We can draw superimpose the nodes on the graph with the iso-lines of the
# function :math:`f`. We can see that the algorithm focuses on the nodes where the
# function has its discontinuities.
cloud_nodes = ot.Cloud(nodes)
cloud_nodes.setPointStyle("dot")
cloud_nodes.setColor("blue")

g.add(cloud_nodes)
g.setLegendPosition("")
view = otv.View(g)

# %%
# Case 2: The integrand function has large gradients.
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# We consider the function:
#
# .. math::
#
#    f : \begin{array}{lcl}
#          \Rset^2 & \rightarrow & \Rset \\
#          (x,y) & \rightarrow & e^{-(x^2+y^2)} + e^{-8((x-4)^2+(y-4)^2)}
#          \end{array}
#
# the domain :math:`\set{D} =[-2,2]\times [-2,2]` and the integral:
#
# .. math::
#    I = \int_{\set{D}} f(x,y)\, dxdy.
#
# Using Maple, we obtain the reference value:
#
# .. math::
#    I_{ref} = 3.51961338289448753812591427600

# %%
# We first define the integrand :math:`f` and the domain :math:`\set{D}`.


def kernel2(x):
    return [
        m.exp(-(x[0] ** 2 + x[1] ** 2))
        + m.exp(-8 * ((x[0] - 4) ** 2 + (x[1] - 4) ** 2))
    ]


integrand = ot.PythonFunction(2, 1, kernel2)
domain = ot.Interval([-2.0, -2.0], [6.0, 6.0])

# %%
# We draw the iso-lines of the integrand function.
g = integrand.draw([-3.0, -3.0], [7.0, 7.0])
g.setTitle(r"Function $f(x,y) = e^{-(x^2+y^2)} + e^{-8((x-4)^2+(y-4)^2)} $")
g.setXTitle("x")
g.setYTitle("y")

# %%
# To get the nodes used to approximate the integral, we need to use a :class:`~openturns.MemoizeFunction`
# that stores all the calls to the function.
integrand_memoized = ot.MemoizeFunction(integrand)
I_value = integration_algo.integrate(integrand_memoized, domain)
print("I = ", I_value[0])
nodes = integrand_memoized.getInputHistory()
print("Nodes : ", nodes)

# %%
# We can draw in blue the nodes on the graph that contains the iso-lines of the
# function :math:`f`. We can see that the algorithm focuses on the nodes where the function
# has fast variations.
cloud_nodes = ot.Cloud(nodes)
cloud_nodes.setPointStyle("dot")
cloud_nodes.setColor("blue")

g.add(cloud_nodes)
g.setLegendPosition("")
view = otv.View(g)


# %%
# Case 3: The integration domain is complex.
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# We consider the function:
#
# .. math::
#
#    f : \begin{array}{lcl}
#          \Rset^2 & \rightarrow & \Rset \\
#          (x,y) & \rightarrow & \cos(2x) \sin(1.5y)
#        \end{array}
#
# the domain :math:`\set{D} = \{(x,y)\, |\, -2\pi \leq x \leq 2\pi, -2 - \cos(x) \leq y  \leq 2 + \cos(x) \}`
# and the integral:
#
# .. math::
#    I = \int_{\set{D}} f(x,y)\, dxdy.
#
# Using Maple, we obtain the reference value:
#
# .. math::
#    I_{ref} = -0.548768615494004063851543284777

# %%
# We first define the integrand :math:`f` and the domain :math:`\set{D}`.


def kernel3(x):
    return [m.cos(2.0 * x[0]) * m.cos(1.5 * x[1])]


integrand = ot.PythonFunction(2, 1, kernel3)

# %%
# The integration domain is defined by the two following functions:
#
# .. math::
#
#    x \rightarrow u(x) = 2 + \cos(x)
#
# .. math::
#
#    x \rightarrow  \ell(x) = -2 - \cos(x)
#
upper_func = ot.SymbolicFunction(["x"], [" 2 + cos(x)"])
lower_func = ot.SymbolicFunction(["x"], ["-2 - cos(x)"])

# %%
# We draw the iso-lines of the integrand function in the domain of integration.
# To do that, we define a new function which is the restriction
# of the integrand function to the integration domain.


def kernel3_insideDomain(x):
    low_x = lower_func([x[0]])[0]
    up_x = upper_func([x[0]])[0]
    if x[1] > low_x and x[1] < up_x:
        return kernel3(x)
    else:
        return [0.0]


integrand_domain = ot.PythonFunction(2, 1, kernel3_insideDomain)
a = 2 * m.pi
b = 4.0
# Here, we increase the default number of levels.
ot.ResourceMap.SetAsUnsignedInteger("Contour-DefaultLevelsNumber", 30)
g = integrand_domain.draw([-a, -b], [a, b])

# %%
# We add the bounds of the domain :math:`\set{D}`.
low_bound = lower_func.draw([-a, -b], [a, b]).getDrawable(0)
up_bound = upper_func.draw([-a, -b], [a, b]).getDrawable(0)
low_bound.setColor("red")
low_bound.setLineWidth(2)
up_bound.setColor("red")
up_bound.setLineWidth(2)

g.add(low_bound)
g.add(up_bound)
g.setTitle(r"Function $f(x,y) =  cos(2x)sin(1.5y)$")
g.setXTitle("x")
g.setYTitle("y")
view = otv.View(g)
# sphinx_gallery_thumbnail_number = 4

# %%
# To get the nodes used to approximate the integral, we need to use a :class:`~openturns.MemoizeFunction`
# that stores all the calls to the function.
integrand_memoized = ot.MemoizeFunction(integrand)
maximumSubIntervals = 4
maximumError = 1e-4
GKRule = ot.GaussKronrodRule(ot.GaussKronrodRule.G7K15)
integration_algo = ot.IteratedQuadrature(
    ot.GaussKronrod(maximumSubIntervals, maximumError, GKRule)
)
I_value = integration_algo.integrate(
    integrand_memoized, -a, a, [lower_func], [upper_func]
)
print("I = ", I_value[0])
nodes = integrand_memoized.getInputHistory()
print("Nodes : ", nodes)

# %%
# We can superimpose the nodes on the graph with the iso-lines of the
# function :math:`f`. We can see that the algorithm focuses on the nodes where
# the function has fast variations.
cloud_nodes = ot.Cloud(nodes)
cloud_nodes.setPointStyle("dot")
cloud_nodes.setColor("black")

g.add(cloud_nodes)
g.setLegendPosition("")
view = otv.View(g)

# %%
# Show all the graphs.
otv.View.ShowAll()



"""
Compute the joint distribution of order statistics
==================================================
"""

# %%
# Context
# -------
#
# In this example, we define the joint distribution of the :math:`n` order statistics
# of a given distribution :math:`X`.
#
# We detail the following cases:
#
# - Case 1: The Min - Max order statistics,
# - Case 2: Two order statistics that bound the :math:`\alpha` -quantile of :math:`X` with the confidence
#   level :math:`1-\beta`.
#
# Let :math:`X` be a random variable and :math:`\vect{X}_{(1:n)} = (X_{(1)},\dots,X_{(n)})`
# be the random vector of all its order statistics  of size :math:`n`.
# Let :math:`f` be its PDF and  :math:`F` its CDF. Then the random vector
# of its  :math:`n` order statistics  is distributed as the random vector
# :math:`(F^{-1}(U_{(1)}), \dots, F^{-1}(U_{(n)}))` where the :math:`U_{(i)}`
# are the order statistics of the Uniform distribution over :math:`[0,1]`:
#
# .. math::
#     (X_{(1)},\dots,X_{(n)}) \sim (F^{-1}(U_{(1))}), \dots, F^{-1}(U_{(n)}))
#
# Then the CDF of :math:`\vect{X}_{(1:n)}` is defined by:
#
# .. math::
#     F_{\vect{X}_{(1:n)}}(\vect{x}) = F_{\vect{U}_{(1:n)}}(F(x_1), \dots, F(x_n))
#
# and its PDF (if defined) by:
#
# .. math::
#    f_{\vect{X}_{(1:n)}}(\vect{x}) = n!\prod_{i=1}^n f(x_i) \,\mathbf{1}_{\cS}(\vect{x})
#    \quad \vect{x} \in \Rset^n
#
# Thus, to get the joint distribution of :math:`\vect{X}_{(1:n)}`, we use the
# :class:`~openturns.JointDistribution` whose all marginals are  :math:`X`
# and whose core is a :class:`~openturns.UniformOrderStatistics`.

# %%
import openturns as ot
import openturns.viewer as otv
import math as m

ot.ResourceMap.SetAsString("Contour-DefaultColorMapNorm", "rank")
# sphinx_gallery_thumbnail_number = 2

# %%
# We create the joint distribution of the order statistics from the distribution of :math:`X`
# and the size of the order statistics:
n = 100
dist = ot.Normal()
orderStat_dist = ot.JointDistribution([dist] * n, ot.UniformOrderStatistics(n))


# %%
# Case 1: The Min - Max order statistics
# --------------------------------------
#
# We extract the first and last marginals of the order statistics distribution:
min_max_dist = orderStat_dist.getMarginal([0, n - 1])


# %%
# We search the mode of the joint distribution. To get it, we search the point that maximizes the
# log-PDF of the distribution. We use a :class:`~openturns.OptimizationProblem`.
optimPb = ot.OptimizationProblem(min_max_dist.getLogPDF())
optimPb.setMinimization(False)
optimPb.setBounds(min_max_dist.getRange())

algo = ot.Cobyla(optimPb)
algo.setStartingPoint(min_max_dist.getMean())
algo.run()

mode = algo.getResult().getOptimalPoint()
print("Mode of the (Min-Max) joint distribution : ", mode)

# %%
# We draw its PDF.
g = min_max_dist.drawPDF()
contour = g.getDrawable(0).getImplementation()
contour.buildDefaultLevels(50)
contour.setIsFilled(True)
g.setDrawable(0, contour)

mode_cloud = ot.Cloud([mode])
mode_cloud.setPointStyle("bullet")
mode_cloud.setColor("red")
mode_cloud.setLegend("mode")
g.add(mode_cloud)

g.setTitle(r"Joint PDF of $(X_{(1)}, X_{(n)})$")
g.setXTitle(r"$X_{(1)}$")
g.setYTitle(r"$X_{(n)}$")
view = otv.View(g)

# %%
# Case 2: Two order statistics that bound the :math:`\alpha`-quantile with the confidence level :math:`1-\beta`.
# ---------------------------------------------------------------------------------------------------------------
#
# We want to bound the  :math:`\alpha`-quantile :math:`q_\alpha` of :math:`X` with two particular order statisctics among the :math:`n` ones.
# We choose these order statisctics  such that they bound  :math:`q_\alpha` with a confidence level equal to
# :math:`1-\beta`.
# For example, we want to bound the quantile of order :math:`90\%` with a confidence  :math:`95\%`.
#
# It is possible to use some particular order statistics to bound :math:`q_\alpha`: [delmas2006]_
# (see proposition 11.1.3) shows that if :math:`i_n` and :math:`j_n` are two indices defined by:
#
# .. math::
#     i_n  = [n \alpha - \sqrt{n} a_{\beta} \sqrt{\alpha(1-\alpha)}]\\
#     j_n  = [n \alpha + \sqrt{n} a_{\beta} \sqrt{\alpha(1-\alpha)}]
#
# where :math:`a_{\beta}` is the quantile of order :math:`(1-\beta/2)` of the normal
# distribution with zero mean and unit variance, then we have:
#
# .. math::
#     \lim_{n \rightarrow +\infty} \Prob{q_\alpha \in [X_{(i_n)}, X_{(j_n)}]} = 1 - \beta
#
# It means that any realization of the joint distribution of :math:`(X_{(i_n)}, X_{(j_n)})`, for :math`n` large enough,
# forms an interval that contains
# :math:`q_\alpha` with a confidence equal to :math:`1-\beta`.
alpha = 0.90
beta = 0.05
a_beta = ot.Normal().computeQuantile(1 - beta / 2)[0]
delta = m.sqrt(n * alpha * (1 - alpha)) * a_beta
i_n = int(n * alpha - delta)
j_n = int(n * alpha + delta)
print("Chosen order statistics (in, jn) = ", i_n, j_n)

# %%
# Be careful that the indices of the order statistics begins at 0.
ic_beta_dist = orderStat_dist.getMarginal([i_n - 1, j_n - 1])

# %%
# We search the mode of the joint distribution. To get it, we search the point that maximizes the
# log-PDF of the distribution. We use a :class:`~openturns.OptimizationProblem`.
optimPb = ot.OptimizationProblem(ic_beta_dist.getLogPDF())
optimPb.setMinimization(False)
optimPb.setBounds(ic_beta_dist.getRange())

algo = ot.Cobyla(optimPb)
algo.setStartingPoint([1.1, 2.1])
algo.run()

mode = algo.getResult().getOptimalPoint()
print("Mode of the joint distribution : ", mode)

# %% We draw the iso-lines of the PDF and the mode.
g = ic_beta_dist.drawPDF()
contour = g.getDrawable(0).getImplementation()
contour.buildDefaultLevels(50)
contour.setIsFilled(True)
g.setDrawable(0, contour)

mode_cloud = ot.Cloud([mode])
mode_cloud.setPointStyle("bullet")
mode_cloud.setColor("red")
mode_cloud.setLegend("mode")
g.add(mode_cloud)

g.setTitle(r"Joint PDF of $(X_{(i_n)}, X_{(j_n)})$")
g.setXTitle(r"$X_{(i_n)}$")
g.setYTitle(r"$X_{(j_n)}$")
view = otv.View(g)

# %%
# Display all figures
otv.View.ShowAll()



"""
Compare frequentist and Bayesian estimation
===========================================
"""

# %%
# In this example, we want to estimate of the parameter
# :math:`\vect{\Theta}` of the distribution of a random vector :math:`\inputRV` from which we have some data.
# We compare the frequentist and the Bayesian approaches to estimate :math:`\vect{\theta}`.
#
# Let :math:`\inputRV = (X_0, X_1)` be a random vector following a bivariate normal distribution
# with zero mean, unit variance and independent components:
# :math:`\cN_2 \left(\vect{\theta}\right)`
# where the parameter :math:`\vect{\theta} = (\mu_0, \sigma_0, \mu_1, \sigma_1, \rho) = (0, 1, 0, 1, 0)`.
# We assume to have a sample generated from :math:`\vect{X}` denoted by
# :math:`(\inputReal_1, \dots, \inputReal_\sampleSize)` where :math:`\sampleSize = 25`.
#
# We assume to know the parameters :math:`(\mu_0, \mu_1, \rho)` and we want to estimate the parameters :math:`(\sigma_0, \sigma_1)`.
# In the Bayesian approach, we assume that :math:`\vect{\Theta} = (0, \Sigma_0, 0, \Sigma_1, 0)` is a random vector and we define
# a link function :math:`g : \Rset^2 \rightarrow \Rset^5` such that:
#
# .. math::
#
#     \vect{\Theta} = g(\vect{Y})
#
# where :math:`\vect{Y}` follows the prior distribution denoted by :math:`\pi_{\vect{Y}}^0`.
#
# Note that the link function :math:`g` already contains the information that the two components of :math:`\inputRV`
# are independent (as :math:`\rho = 0`) and centered (as :math:`\mu_i = 0`).
#
# We consider two interesting link functions:
#
# .. math::
#
#   g_1(\vect{y}) & = (0, y_0, 0, y_1, 0) \\
#   g_2(\vect{y}) & = (0,0.5+y_0^2, 0, 0.5+y_1^2, 0)
#
# each one being associated to the respective prior distributions:
#
# .. math::
#
#   \pi_{\vect{Y}}^{0,1} & = \cT(0,1,2) \times \cT(0,1,2) \\
#   \pi_{\vect{Y}}^{0,2} & = \cT(-1, 0, 1) \times \cT(-1, 0,1)
#
# The second case is such that the link function :math:`g_2` is not bijective on the range of :math:`\pi_{\vect{Y}}^{0,2}`.
#
# The Bayesian approach uses the :class:`~openturns.PosteriorDistribution` that estimates the posterior distribution of :math:`\vect{Y}` denoted by
# :math:`\pi_{\vect{Y}}^\sampleSize` maximizing the likelihood of the conditioned model on the sample, weighted by the prior distribution
# :math:`\pi_{\vect{Y}}^0`. From the :math:`\pi_{\vect{Y}}^\sampleSize` distribution, we extract the vector of modes
# denoted by :math:`\vect{Y}_n^m`: this point maximizes :math:`\pi_{\vect{Y}}^\sampleSize`.
#
# It is interesting to note that when :math:`n \rightarrow +\infty`, then :math:`\pi_{\vect{Y}}^\sampleSize \rightarrow \pi_{\vect{Y}}^*`
# such that :math:`g(\pi_{\vect{Y}}^*)` is a Dirac distribution centered at :math:`\vect{\theta}`.
#
# The frequentist approach estimates the parameter :math:`\vect{\theta}` by maximizing the likelihood of the normal model on the
# sample. To model the same level of information, we consider a centered bivariate normal model with independent components.
# We denote by :math:`\vect{\theta}_n^{MC}` the parameter obtained.


# %%
import openturns as ot
import openturns.viewer as otv

ot.ResourceMap.SetAsUnsignedInteger(
    "CompoundDistribution-MarginalIntegrationNodesNumber", 32
)
ot.ResourceMap.SetAsString(
    "CompoundDistribution-ContinuousDiscretizationMethod", "GaussProduct"
)


# %%
# Here, we define a function that computes the mode of a distribution, which is the point that maximises its PDF.


def computeMode(distribution):
    def obj_py(X):
        return distribution.computeLogPDF(X) * (-1.0)

    obj = ot.PythonFunction(distribution.getDimension(), 1, func_sample=obj_py)
    pb = ot.OptimizationProblem(obj)
    pb.setBounds(distribution.getRange())
    algo = ot.Cobyla(pb)
    algo.setStartingPoint(distribution.getMean())
    algo.run()
    return algo.getResult().getOptimalPoint()


# %%
# Define the theoretical :math:`\vect{X}` distribution which is a normal distribution with zero mean,
# unit variance and independent components.
conditioned = ot.Normal(2)
conditioned.setDescription(["X0", "X1"])

# %%
# We create the sample that we are going to use to estimate the parameters.
Nsample = 25
observations = conditioned.getSample(Nsample)


# %%
# Case 1: Bijective link function
# -------------------------------
#
# Here, :math:`g_1(\vect{y}) = (0, y_0, 0, y_1, 0)` and :math:`\pi_{\vect{Y}}^{0,1} = \cT(0,1,2) \times \cT(0,1,2)`.
#
# We create the link function.
linkFunction = ot.SymbolicFunction(["u0", "u1"], ["0.0", "u0", "0.0", "u1", "0.0"])

# %%
# We create the prior distribution of :math:`\vect{Y}`.
conditioning = ot.JointDistribution([ot.Triangular(0.0, 1.0, 2.0)] * 2)
conditioning.setDescription(["Y0", "Y1"])

# %%
# We have to decondition the :math:`\vect{X}|\vect{\Theta} = g(\vect{Y})` distribution with
# respect to the  prior distribution :math:`\pi_{\vect{Y}}^{0,1}` in order to get the
# final distribution of :math:`\vect{X}`. To do that, we use the :class:`~openturns.CompoundDistribution`.
compound = ot.CompoundDistribution(conditioned, conditioning, linkFunction)

# %%
# Then, we can create the posterior distribution :math:`\pi_{\vect{Y}}^\sampleSize` based on the compound distribution of :math:`\vect{X}` and
# the sample.
posterior_Y = ot.PosteriorDistribution(compound, observations)

# %%
# From  :math:`\pi_{\vect{Y}}^\sampleSize`, we get:
#
# - the mode parameter :math:`\vect{Y}_n^m` that maximizes the PDF,
# - the distribution of :math:`\inputRV`  parameterized by :math:`\vect{\theta}_n^m = g(\vect{Y}_n^m)`,
# - a bilateral condifence interval of level :math:`\alpha` of the posterior distribution of :math:`\vect{Y}`,
# - the volume of this condifence interval,
# - the Kullback-Leibler distance between the Bayesian mode based distribution and the theoretical one:
#   :math:`KL\left(\cN_2\left(g(\vect{Y}_n^m)\right), \cN_2 \left(\vect{\theta}\right) \right)`.
#
theta_Bay = linkFunction(computeMode(posterior_Y))
print("Theta Bay =", theta_Bay)
model_Bay = ot.Distribution(conditioned)
model_Bay.setParameter(theta_Bay)
dist_estimateur_Bay = posterior_Y
alpha = 0.95
interval_Bay, beta = (
    dist_estimateur_Bay.computeBilateralConfidenceIntervalWithMarginalProbability(alpha)
)
print("Beta =", beta)
print("Condifence interval Bay =\n", interval_Bay)
print("Volume =", interval_Bay.getVolume())
sample = model_Bay.getSample(10000)
dist_Bay = (
    model_Bay.computeLogPDF(sample) - conditioned.computeLogPDF(sample)
).computeMean()
print("Kullback-Leibler distance Bay =", dist_Bay[0])


# %%
# The frequentist approach estimates a normal distribution from the sample. We fix the known parameters :math:`(\mu_0, \mu_1, \rho)`
# to their true values.
lh_factory = ot.NormalFactory()
lh_factory.setKnownParameter([0, 2, 4], [0.0] * 3)
lh_est = lh_factory.buildEstimator(observations)

# %%
# We extract from the likelihood estimate:
#
# - the asymptotic distribution of the estimator,
# - the parameters estimates,
# - a bilateral condifence interval of level :math:`\alpha` of the asymptotic distribution of the estimator,
# - the volume of this condifence interval,
# - the Kullback-Leibler distance between the maximum likelihood based distribution and the theoretical one:
#   :math:`KL\left(\cN_2 \left(\vect{\theta}_n^{MV} \right), \cN_2 \left(\vect{\theta}\right) \right)`.
#
model_ML = lh_est.getDistribution()
theta_ML = model_ML.getParameter()
print("Theta ML = ", theta_Bay)
dist_estimator_ML = lh_est.getParameterDistribution().getMarginal([1, 3])
interval_ML, beta = (
    dist_estimator_ML.computeBilateralConfidenceIntervalWithMarginalProbability(alpha)
)
print("Beta =", beta)
print("Condifence interval ML =\n", interval_ML)
print("Volume =", interval_ML.getVolume())
sample = model_ML.getSample(10000)
dist_KL = (
    model_ML.computeLogPDF(sample) - conditioned.computeLogPDF(sample)
).computeMean()
print("Kullback-Leibler distance ML =", dist_KL[0])


# %%
# In the following figure, we  plot:
#
# - the theoretical distribution of :math:`\inputRV`:  :math:`\cN_2 \left(\vect{\theta}\right)` (solid lines),
# - its maximum likelihood based distribution: :math:`\cN_2 \left(\vect{\theta}_n^{MV}\right)` (dashed lines),
# - its Bayesian mode based distribution: :math:`\cN_2(g(\vect{Y}_n^m))` (dotted lines).
#
# We conclude that both approaches lead to the same results.
ot.ResourceMap.SetAsString("Contour-DefaultColorMapNorm", "rank")
g = conditioned.drawPDF()
levels = g.getDrawable(0).getLevels()
dr_ML = model_ML.drawPDF().getDrawable(0).getImplementation()
dr_ML.setLevels(levels)
dr_ML.setColorBarPosition("")
dr_ML.setLineStyle("dashed")
g.add(dr_ML)
dr_Bay = model_Bay.drawPDF().getDrawable(0).getImplementation()
dr_Bay.setLevels(levels)
dr_Bay.setColorBarPosition("")
dr_Bay.setLineStyle("dotted")
g.add(dr_Bay)
g.add(ot.Cloud(observations))
g.setLegends(["Theoretical dist", "ML dist", "Bay dist", "Observations"])
g.setXTitle(r"$X_0$")
g.setYTitle(r"$X_1$")
g.setTitle("Initial distribution, ML estimated dist and Bayesian estimated dist.")
view = otv.View(g, (800, 800), square_axes=True)

# %%
# In the following figure, we consider the parameter :math:`\vect{\Theta}` and we plot:
#
# - the asymptotic distribution of the maximum likelihood estimator: :math:`\vect{\theta}_n^{MV}` (left),
# - the Bayesian mode based distribution: :math:`g(\pi_{\vect{Y}}^\sampleSize)` (right).
#
# On each figure, we draw the bilateral confidence interval or bilateral credibility interval of
# level :math:`\alpha = 0.95`
# computed from the estimator distribution.
#
# First the maximum likelihood estimator.
g_ML = dist_estimator_ML.drawPDF([0.5] * 2, [1.5] * 2)
c = ot.Cloud([theta_ML[[1, 3]]])
c.setColor("red")
c.setPointStyle("bullet")
g_ML.add(c)
c = ot.Cloud([theta_Bay[[1, 3]]])
c.setColor("red")
c.setPointStyle("+")
g_ML.add(c)
a = interval_ML.getLowerBound()
b = interval_ML.getUpperBound()
c = ot.Curve([a, [a[0], b[1]], b, [b[0], a[1]], a])
c.setColor("blue")
g_ML.add(c)
a = interval_Bay.getLowerBound()
b = interval_Bay.getUpperBound()
c = ot.Curve([a, [a[0], b[1]], b, [b[0], a[1]], a])
c.setColor("blue")
c.setLineStyle("dashed")
g_ML.add(c)
g_ML.setLegends(
    [
        r"dist of $\mathbf{\theta}_n^{MV}$",
        r"$\mathbf{\theta}_n^{MV}$",
        r"$g(\mathbf{Y}_n^m)$",
        "ML CI(" + str(int(100 * alpha)) + "%)",
        "Bay CI(" + str(int(100 * alpha)) + "%)",
    ]
)
g_ML.setXTitle(r"$\sigma_0$")
g_ML.setYTitle(r"$\sigma_1$")
g_ML.setTitle("ML Estimator")

# %%
# Then the Bayesian estimator.
g_Bay = dist_estimateur_Bay.drawPDF([0.5] * 2, [1.5] * 2)
c = ot.Cloud([theta_Bay[[1, 3]]])
c.setColor("red")
c.setPointStyle("bullet")
g_Bay.add(c)
c = ot.Cloud([theta_ML[[1, 3]]])
c.setColor("red")
c.setPointStyle("+")
g_Bay.add(c)
a = interval_Bay.getLowerBound()
b = interval_Bay.getUpperBound()
c = ot.Curve([a, [a[0], b[1]], b, [b[0], a[1]], a])
c.setColor("blue")
g_Bay.add(c)
a = interval_ML.getLowerBound()
b = interval_ML.getUpperBound()
c = ot.Curve([a, [a[0], b[1]], b, [b[0], a[1]], a])
c.setColor("blue")
c.setLineStyle("dashed")
g_Bay.add(c)
g_Bay.setLegends(
    [
        r"dist of $g(\pi_{n, \mathbf{Y}})$",
        r"$g(\mathbf{Y}_n^m)$",
        r"$\mathbf{\theta}_n^{MV}$",
        "Bay CI(" + str(int(100 * alpha)) + "%)",
        "ML CI(" + str(int(100 * alpha)) + "%)",
    ]
)
g_Bay.setXTitle(r"$\sigma_0$")
g.setYTitle(r"$\sigma_1$")
g_Bay.setTitle("Bayesian Estimator")

# Gather both graph in a grid.
grid = ot.GridLayout(1, 2)
grid.setGraph(0, 0, g_ML)
grid.setGraph(0, 1, g_Bay)

view = otv.View(grid)

# %%
# Finally, the following table sums up the previous computed quantities.
#
# ===========      ========================  ========================  =====================    =====================
# Approach         :math:`\tilde{\sigma}_0`  :math:`\tilde{\sigma}_1`  IC(:math:`\sigma_0`)     IC(:math:`\sigma_1`)
# ===========      ========================  ========================  =====================    =====================
# Frequentist      0.967                     0.951                     :math:`[0.651, 1.30]`    :math:`[0.659, 1.20]`
# Bayesian         0.950                     0.934                     :math:`[0.736, 1.38]`    :math:`[0.724, 1.35]`
# ===========      ========================  ========================  =====================    =====================
#
# ===========      =============   ======================   =================================================
# Approach         :math:`\beta`   :math:`KL`               :math:`\cV(IC(\tilde{\sigma}_0, \tilde{\sigma}))`
# ===========      =============   ======================   =================================================
# Frequentist      0.974           :math:`3.46\, 10^{-3}`   0.352
# Bayesian         0.974           :math:`6.87\, 10^{-3}`   0.381
# ===========      =============   ======================   =================================================


# %%
# Case 2: we consider the second link function
# --------------------------------------------
#
# Here, :math:`g_2(\vect{y}) = (0,0.5+y_0^2, 0, 0.5+y_1^2, 0)` and :math:`\pi_{\vect{Y}}^{0,2} = \cT(-1, 0, 1) \times \cT(-1, 0,1)`.
#
# We note that the posterior distribution of :math:`\vect{Y}` is quadri-modal, as the link function :math:`g_2` is no more bijective
# on the range of :math:`\pi_{\vect{Y}}^{0,2}`.
#
# We go through the same steps as described previously. The maximum estimator is not changed.
# But in order to get the posterior distribution of :math:
linkFunction = ot.SymbolicFunction(
    ["u0", "u1"], ["0.0", "0.5+u0^2", "0.0", "0.5+u1^2", "0.0"]
)
conditioning = ot.JointDistribution([ot.Triangular(-1.0, 0.0, 1.0)] * 2)
conditioning.setDescription(["Y0", "Y1"])
compound = ot.CompoundDistribution(conditioned, conditioning, linkFunction)
posterior_Y = ot.PosteriorDistribution(compound, observations)
sample_posterior = linkFunction(posterior_Y.getSample(10000)).getMarginal([1, 3])
dist_estimateur_Bay = ot.KernelSmoothing().build(sample_posterior)

theta_Bay = linkFunction(computeMode(posterior_Y))
print("Theta Bay =", theta_Bay)
model_Bay = ot.Distribution(conditioned)
model_Bay.setParameter(theta_Bay)
alpha = 0.95
interval_Bay, beta = (
    dist_estimateur_Bay.computeBilateralConfidenceIntervalWithMarginalProbability(alpha)
)
print("Beta =", beta)
print("Condifence interval Bay =\n", interval_Bay)
print("Volume =", interval_Bay.getVolume())
sample = model_Bay.getSample(10000)
dist_Bay = (
    model_Bay.computeLogPDF(sample) - conditioned.computeLogPDF(sample)
).computeMean()
print("Kullback-Leibler distance Bay =", dist_Bay[0])

# %%
g = conditioned.drawPDF()
levels = g.getDrawable(0).getLevels()
dr_ML = model_ML.drawPDF().getDrawable(0).getImplementation()
dr_ML.setLevels(levels)
dr_ML.setColorBarPosition("")
dr_ML.setLineStyle("dashed")
g.add(dr_ML)
dr_Bay = model_Bay.drawPDF().getDrawable(0).getImplementation()
dr_Bay.setLevels(levels)
dr_Bay.setColorBarPosition("")
dr_Bay.setLineStyle("dotted")
g.add(dr_Bay)
g.add(ot.Cloud(observations))
g.setLegends(["Theoretical dist", "ML dist", "Bay dist", "Observations"])
g.setXTitle(r"$X_0$")
g.setYTitle(r"$X_1$")
g.setTitle("Initial distribution, ML estimated dist and Bayesian estimated dist.")
view = otv.View(g, (800, 800), square_axes=True)

# %%
g_ML = dist_estimator_ML.drawPDF([0.5] * 2, [1.5] * 2)
c = ot.Cloud([theta_ML[[1, 3]]])
c.setColor("red")
c.setPointStyle("bullet")
g_ML.add(c)
c = ot.Cloud([theta_Bay[[1, 3]]])
c.setColor("red")
c.setPointStyle("+")
g_ML.add(c)
a = interval_ML.getLowerBound()
b = interval_ML.getUpperBound()
c = ot.Curve([a, [a[0], b[1]], b, [b[0], a[1]], a])
c.setColor("blue")
g_ML.add(c)
a = interval_Bay.getLowerBound()
b = interval_Bay.getUpperBound()
c = ot.Curve([a, [a[0], b[1]], b, [b[0], a[1]], a])
c.setColor("blue")
c.setLineStyle("dashed")
g_ML.add(c)
g_ML.setLegends(
    [
        r"dist of $\mathbf{\theta}_n^{MV}$",
        r"$\mathbf{\theta}_n^{MV}$",
        r"$g(\mathbf{Y}_n^m)$",
        "ML CI(" + str(int(100 * alpha)) + "%)",
        "Bay CI(" + str(int(100 * alpha)) + "%)",
    ]
)
g_ML.setXTitle(r"$\sigma_0$")
g_ML.setYTitle(r"$\sigma_1$")
g_ML.setTitle("ML Estimator")


g_Bay = dist_estimateur_Bay.drawPDF([0.5] * 2, [1.5] * 2)
c = ot.Cloud([theta_Bay[[1, 3]]])
c.setColor("red")
c.setPointStyle("bullet")
g_Bay.add(c)
c = ot.Cloud([theta_ML[[1, 3]]])
c.setColor("red")
c.setPointStyle("+")
g_Bay.add(c)
a = interval_Bay.getLowerBound()
b = interval_Bay.getUpperBound()
c = ot.Curve([a, [a[0], b[1]], b, [b[0], a[1]], a])
c.setColor("blue")
g_Bay.add(c)
a = interval_ML.getLowerBound()
b = interval_ML.getUpperBound()
c = ot.Curve([a, [a[0], b[1]], b, [b[0], a[1]], a])
c.setColor("blue")
c.setLineStyle("dashed")
g_Bay.add(c)
g_Bay.setLegends(
    [
        r"dist of $g(\pi_{n, \mathbf{Y}})$",
        r"$g(\mathbf{Y}_n^m)$",
        r"$\mathbf{\theta}_n^{MV}$",
        "Bay CI(" + str(int(100 * alpha)) + "%)",
        "ML CI(" + str(int(100 * alpha)) + "%)",
    ]
)
g_Bay.setXTitle(r"$\sigma_0$")
g.setYTitle(r"$\sigma_1$")
g_Bay.setTitle("Bayesian Estimator")

grid = ot.GridLayout(1, 2)
grid.setGraph(0, 0, g_ML)
grid.setGraph(0, 1, g_Bay)

view = otv.View(grid)

# %%
# Finally, the following table sums up the previous computed quantities.
#
# ===========      ========================  ========================  =====================    =====================
# Approach         :math:`\tilde{\sigma}_0`  :math:`\tilde{\sigma}_1`  IC(:math:`\sigma_0`)     IC(:math:`\sigma_1`)
# ===========      ========================  ========================  =====================    =====================
# Frequentist      0.967                     0.951                     :math:`[0.651, 1.30]`    :math:`[0.659, 1.20]`
# Bayesian         0.929                     0.915                     :math:`[0.706, 1.27]`    :math:`[0.693, 1.25]`
# ===========      ========================  ========================  =====================    =====================
#
# ===========      =============   ======================   =================================================
# Approach         :math:`\beta`   :math:`KL`               :math:`\cV(IC(\tilde{\sigma}_0, \tilde{\sigma}))`
# ===========      =============   ======================   =================================================
# Frequentist      0.974           :math:`3.46\, 10^{-3}`   0.352
# Bayesian         0.974           :math:`1.21\, 10^{-2}`   0.316
# ===========      =============   ======================   =================================================
#
# We also plot the PDF of the posterior distribution :math:`\pi_{\vect{Y}}^\sampleSize` of :math:`\vect{Y}` which is quadri-modal, with a sample.
# sphinx_gallery_thumbnail_number =  5
g_pinY = posterior_Y.drawPDF()
g_pinY.setXTitle(r"$Y_0$")
g_pinY.setYTitle(r"$Y_1$")
g_pinY.setTitle(r"Posterior Bayesian Distribution of $\mathbf{Y}$")
g_pinY.add(ot.Cloud(posterior_Y.getSample(100)))
view = otv.View(g_pinY, (800, 800))

# %%
view.ShowAll()



"""
Use case : vertical deflection of a tube
========================================
"""

import openturns as ot


class DeflectionTube:
    r"""
    Data class for the deflection of a tube model.


    Attributes
    ----------

    dim : int
        The dimension of the problem: dim=6

    model : :class:`~openturns.SymbolicFunction`
        Model of the deflection.
        The model has input dimension 6 and output dimension 3.
        More precisely, we have :math:`\vect{X} = (F, L, a, D, d, E)`
        and  :math:`Y = (y, \theta_L, \theta_R)`.

    XF : :class:`~openturns.Normal`
        F distribution
        `Normal(1, 0.1)`

    XE : :class:`~openturns.Normal`
        E distribution
        `Normal(200000, 2000)`

    XL : :class:`~openturns.Dirac`
        L distribution
        `Dirac(1.5)`

    Xa : :class:`~openturns.Dirac`
        a distribution
        `Dirac(1.0)`

    XD : :class:`~openturns.Dirac`
        D distribution
        `Dirac(0.8)`

    Xd : :class:`~openturns.Dirac`
        d distribution
        `Dirac(0.1)`

    distribution : :class:`~openturns.JointDistribution`
        The joint distribution of the input parameters.

    Examples
    --------
    >>> from openturns.usecases import deflection_tube
    >>> # Load the deflection tube model
    >>> dt = deflection_tube.DeflectionTube()
    >>> print("Inputs:", dt.model.getInputDescription())
    Inputs: [F,L,a,De,di,E]
    >>> print("Outputs:", dt.model.getOutputDescription())
    [Deflection,Left angle,Right angle]
    """

    def __init__(self):
        self.dim = 6

        formula = "var I := pi_*(De^4 - di^4) / 32;"
        formula += "var b:=L - a; g1 := -F * a^2 * (L - a)^2 / (3 * E * L * I);"
        formula += "g2 := -F * b * (L^2 - b^2) / (6 * E * L * I);"
        formula += "g3 := F * a * (L^2 - a^2) / (6 * E * L * I)"
        self.model = ot.SymbolicFunction(
            ["F", "L", "a", "De", "di", "E"],
            ["g1", "g2", "g3"],
            formula,
        )
        self.model.setOutputDescription(["Deflection", "Left angle", "Right angle"])

        self.XF = ot.Normal(1, 0.1)
        self.XF.setDescription(["Force"])

        self.XE = ot.Normal(200000, 2000)
        self.XE.setDescription(["Young Modulus"])

        self.XL = ot.Dirac(1.5)
        self.XL.setDescription(["Length"])

        self.Xa = ot.Dirac(1.0)
        self.Xa.setDescription(["Location"])

        self.XD = ot.Dirac(0.8)
        self.XD.setDescription(["External diameter"])

        self.Xd = ot.Dirac(0.1)
        self.Xd.setDescription(["Internal diameter"])

        self.distribution = ot.JointDistribution(
            [self.XF, self.XL, self.Xa, self.XD, self.Xd, self.XE]
        )
        self.inputDistribution = self.distribution  # deprecated



#! /usr/bin/env python

import openturns as ot

ot.TESTPREAMBLE()
ot.ResourceMap.SetAsUnsignedInteger(
    "LinearCombinationDistribution-DefaultMaxSize", 4000000
)
ot.TTY.ShowColors(False)

# Test fusion of Dirac with no other atoms: should be a unique Dirac(5.8)
d = ot.LinearCombinationDistribution(
    [ot.Dirac(1.0), ot.Dirac(2.0), ot.Dirac(3.0)], [0.5, 0.6, 0.7], 2.0
)
print("d=", d)
# Test fusion of Dirac with other atoms: the Dirac should be merged into
# the constant: 5 + Exponential(lambda=1,gamma=0)
d = ot.LinearCombinationDistribution(
    [ot.Dirac(1.0), ot.Dirac(2.0), ot.Exponential()], 2.0
)
print("d=", d)
# Test flatten LinearCombinationDistribution atoms: the LinearCombinationDistribution should have 4 atoms.
d = ot.LinearCombinationDistribution(
    [
        ot.Gumbel(),
        ot.LinearCombinationDistribution(
            [ot.Logistic(), ot.WeibullMin()], [0.5, 1.5], 3.0
        ),
        ot.Frechet(1.0, 4.0),
    ],
    [2.0, 3.0, 4.0],
    2.0,
)
print("d=", d)
# Test merge of Normal atoms:
d = ot.LinearCombinationDistribution(
    [ot.Normal(1.0, 8.0), ot.Logistic(), ot.Normal(2.0, 1.0)], [0.5, 2.5, 3.0], 2.0
)
print("d=", d)
# Test merge of Exponential, Gamma and ChiSquare atoms
d = ot.LinearCombinationDistribution(
    [
        ot.Exponential(1.0, 1.0),
        ot.Exponential(1.5, -1.0),
        ot.Exponential(2.0, 1.0),
        ot.Gamma(4.0, 2.0, -1.0),
        ot.Gamma(3.0, 1.0, 3.0),
        ot.ChiSquare(4.0),
    ],
    [1.0, 1.5, 1.0, 2.0, 2.0, 0.5],
    2.0,
)
print("d=", d)
# Test merge of Uniform atoms
d = ot.LinearCombinationDistribution(
    [
        ot.Uniform(0.0, 1.0),
        ot.Uniform(0.0, 1.0),
        ot.Uniform(1.0, 3.0),
        ot.Uniform(-1.0, 4.0),
        ot.Uniform(2.0, 3.0),
    ],
    [1.0, 1.0, 2.0, 3.0, 4.0],
    2.0,
)
print("d=", d)
# Test merge of Bernoulli and Binomial atoms
# Deactivate the fusion of discrete atoms
ot.ResourceMap.SetAsUnsignedInteger(
    "LinearCombinationDistribution-MaximumSupportSize", 0
)
d = ot.LinearCombinationDistribution(
    [
        ot.Bernoulli(0.5),
        ot.Bernoulli(0.5),
        ot.Bernoulli(0.1),
        ot.Binomial(4, 0.5),
        ot.Binomial(6, 0.5),
        ot.Binomial(3, 0.1),
    ],
    [1.0, 1.5, 2.0, 1.0, 4.0, 2.0],
    2.0,
)
print("d=", d)
# Test merge of Poisson atoms
# Deactivate the fusion of discrete atoms
ot.ResourceMap.SetAsUnsignedInteger(
    "LinearCombinationDistribution-MaximumSupportSize", 0
)
d = ot.LinearCombinationDistribution(
    [
        ot.Poisson(3.0),
        ot.Poisson(2.0),
        ot.Poisson(6.0),
        ot.Poisson(10.0),
        ot.Poisson(4.0),
    ],
    [1.0, 2.0, 3.0, 2.0, 1.0],
    2.0,
)
print("d=", d)
# Test merge of discrete and continuous atoms
# Deactivate the fusion of discrete atoms
ot.ResourceMap.SetAsUnsignedInteger(
    "LinearCombinationDistribution-MaximumSupportSize", 0
)
# more continuous atoms
d = ot.LinearCombinationDistribution(
    [ot.Logistic(), ot.Binomial(2, 0.5), ot.Uniform()], [1.0, 2.0, 3.0], 2.0
)
print("d=", d)
# more discrete atoms
d = ot.LinearCombinationDistribution(
    [ot.Bernoulli(0.1), ot.Binomial(2, 0.5), ot.Uniform()], [1.0, 2.0, 3.0], 2.0
)
print("d=", d)
# same number of continuous and discrete atoms
d = ot.LinearCombinationDistribution(
    [ot.Logistic(), ot.Bernoulli(0.1), ot.Binomial(2, 0.5), ot.Uniform()],
    [1.0, 2.0, 3.0, 4.0],
    2.0,
)
print("d=", d)
# Test the fusion of discrete atoms
ot.Log.Show(ot.Log.DBG)
print("\n")
# All the atoms have a too large support
ot.ResourceMap.SetAsUnsignedInteger(
    "LinearCombinationDistribution-MaximumSupportSize", 1
)
d = ot.LinearCombinationDistribution(
    [ot.Binomial(2, 0.1), ot.Binomial(3, 0.5), ot.Poisson(), ot.Geometric()],
    [1.0, 2.0, 3.0, 4.0],
    2.0,
)
print("d=", d)
print("\n")
# Some atoms have a too large support, no pending aggregated discrete
ot.ResourceMap.SetAsUnsignedInteger(
    "LinearCombinationDistribution-MaximumSupportSize", 15
)
d = ot.LinearCombinationDistribution(
    [ot.Binomial(2, 0.1), ot.Binomial(3, 0.5), ot.Poisson(), ot.Geometric()],
    [1.0, 2.0, 3.0, 4.0],
    2.0,
)
print("d=", d)
print("\n")
# Some atoms have a too large support, a pending aggregated discrete
ot.ResourceMap.SetAsUnsignedInteger(
    "LinearCombinationDistribution-MaximumSupportSize", 15
)
d = ot.LinearCombinationDistribution(
    [
        ot.Binomial(2, 0.1),
        ot.Binomial(3, 0.5),
        ot.Poisson(),
        ot.Binomial(2, 0.1),
        ot.Binomial(3, 0.5),
    ],
    [1.0, 2.0, 3.0, 4.0, 5.0],
    2.0,
)
print("d=", d)
print("\n")
# All the atoms can be merged
ot.ResourceMap.SetAsUnsignedInteger(
    "LinearCombinationDistribution-MaximumSupportSize", 100
)
d = ot.LinearCombinationDistribution(
    [ot.Bernoulli(0.1), ot.Bernoulli(0.2), ot.Bernoulli(0.3), ot.Bernoulli(0.4)],
    [1.0, 2.0, 3.0, 4.0],
    2.0,
)
print("d=", d)



#! /usr/bin/env python

import openturns as ot

ot.TESTPREAMBLE()


size = 10
dimension = 2
sample = ot.Sample(size, dimension)
# Fill-in the sample
for i in range(size):
    p = ot.Point(dimension)
    for j in range(dimension):
        p[j] = i + j * 1.0 / dimension
    sample[i] = p
print("sample=", repr(sample))
# History using the Null strategy
nullStrategy = ot.Null()
for i in range(size):
    nullStrategy.store(sample[i])
print("Null strategy sample=", repr(nullStrategy.getSample()))
# History using the Full strategy
fullStrategy = ot.Full()
fullStrategy.setDimension(dimension)
for i in range(size):
    fullStrategy.store(sample[i])
print("Full strategy sample=", repr(fullStrategy.getSample()))
# History using the Last strategy, large storage
lastStrategy = ot.Last(3 * size)
lastStrategy.setDimension(dimension)
for i in range(size):
    lastStrategy.store(sample[i])
print("Last strategy sample (large storage)=", repr(lastStrategy.getSample()))
lastStrategy = ot.Last(size // 3)
lastStrategy.setDimension(dimension)
# History using the Last strategy, small storage
for i in range(size):
    lastStrategy.store(sample[i])
print("Last strategy sample (small storage)=", repr(lastStrategy.getSample()))
# History using the Compact strategy, large storage
compactStrategy = ot.Compact(3 * size)
compactStrategy.setDimension(dimension)
for i in range(size):
    compactStrategy.store(sample[i])
print("Compact strategy sample (large storage)=", repr(compactStrategy.getSample()))
compactStrategy = ot.Compact(size // 3)
compactStrategy.setDimension(dimension)
# History using the Compact strategy, small storage
for i in range(size):
    compactStrategy.store(sample[i])
print("Compact strategy sample (small storage)=", repr(compactStrategy.getSample()))



#! /usr/bin/env python
import openturns as ot
import math
import openturns.testing as ott

ot.TESTPREAMBLE()

f = ot.SymbolicFunction(["x"], ["sin(x)"])
a = -2.5
b = 4.5
# Integrate sin(t) between a & b --> cos(b) - sin(b)
ref = math.cos(a) - math.cos(b)

all_methods = [
    ot.FejerAlgorithm.FEJERTYPE1,
    ot.FejerAlgorithm.FEJERTYPE2,
    ot.FejerAlgorithm.CLENSHAWCURTIS,
]
# 1D checking
for method in all_methods:
    algo = ot.FejerAlgorithm([100], method)
    value, adaptedNodes = algo.integrateWithNodes(f, ot.Interval(a, b))
    ott.assert_almost_equal(value[0], ref, 1e-10, 1e-10)

g = ot.SymbolicFunction(["x", "y"], ["cos(pi_ * x / 2) * sin(pi_ * y)"])
ref = 8 / (math.pi * math.pi)
interval = ot.Interval([-1, 0], [1, 1])
for method in all_methods:
    algo = ot.FejerAlgorithm([64, 64], method)
    value, adaptedNodes = algo.integrateWithNodes(g, interval)
    ott.assert_almost_equal(value[0], ref, 1e-10, 1e-10)

# Now we use the same calculus using variables changes
h = ot.SymbolicFunction(
    ["x", "y"], ["cos(pi_ * x / 2) * sin(pi_ * y / 2 + pi_/2 ) / 2"]
)
interval = ot.Interval([-1, -1], [1, 1])
for method in all_methods:
    algo = ot.FejerAlgorithm([64, 64], method)
    value, adaptedNodes = algo.integrateWithNodes(h, interval)
    ott.assert_almost_equal(value[0], ref, 1e-10, 1e-10)

for n in range(1, 8):
    print(f"-- n={n} {'-' * 50}")
    for rule in all_methods:
        algo = ot.FejerAlgorithm([n], rule)
        nodes = algo.getNodes().asPoint()
        weights = algo.getWeights()
        print(f"rule={rule}", nodes, weights)
        if n == 1 and rule == ot.FejerAlgorithm.CLENSHAWCURTIS:
            ott.assert_almost_equal(nodes, [0.0])
        if n == 2 and rule == ot.FejerAlgorithm.FEJERTYPE2:
            ott.assert_almost_equal(weights, [1.0] * 2)



"""
Sobol' sensitivity indices from chaos
=====================================
"""

# %%
# In this example we are going to compute global sensitivity indices from a functional chaos decomposition.
#
# We study the Borehole function that models water flow through a borehole:
#
# .. math::
#    \frac{2 \pi T_u (H_u - H_l)}{\ln{r/r_w}(1+\frac{2 L T_u}{\ln(r/r_w) r^2_w K_w}\frac{T_u}{T_l})}
#
# With parameters:
#
# - :math:`r_w`: radius of borehole (m)
# - :math:`r`: radius of influence (m)
# - :math:`T_u`: transmissivity of upper aquifer (:math:`m^2/yr`)
# - :math:`H_u`: potentiometric head of upper aquifer (m)
# - :math:`T_l`: transmissivity of lower aquifer (:math:`m^2/yr`)
# - :math:`H_l`: potentiometric head of lower aquifer (m)
# - :math:`L`: length of borehole (m)
# - :math:`K_w`: hydraulic conductivity of borehole (:math:`m/yr`)

# %%
import openturns as ot
from operator import itemgetter
import openturns.viewer as otv


# %%
# borehole model
dimension = 8
input_names = ["rw", "r", "Tu", "Hu", "Tl", "Hl", "L", "Kw"]
model = ot.SymbolicFunction(
    input_names, ["(2*pi_*Tu*(Hu-Hl))/(ln(r/rw)*(1+(2*L*Tu)/(ln(r/rw)*rw^2*Kw)+Tu/Tl))"]
)
coll = [
    ot.Normal(0.1, 0.0161812),
    ot.LogNormal(7.71, 1.0056),
    ot.Uniform(63070.0, 115600.0),
    ot.Uniform(990.0, 1110.0),
    ot.Uniform(63.1, 116.0),
    ot.Uniform(700.0, 820.0),
    ot.Uniform(1120.0, 1680.0),
    ot.Uniform(9855.0, 12045.0),
]
distribution = ot.JointDistribution(coll)
distribution.setDescription(input_names)

# %%
# Freeze r, Tu, Tl from model to go faster
selection = [1, 2, 4]
complement = ot.Indices(selection).complement(dimension)
distribution = distribution.getMarginal(complement)
model = ot.ParametricFunction(
    model, selection, distribution.getMarginal(selection).getMean()
)
input_names_copy = list(input_names)
input_names = itemgetter(*complement)(input_names)
dimension = len(complement)

# %%
# design of experiment
size = 1000
X = distribution.getSample(size)
Y = model(X)

# %%
# create a functional chaos model
algo = ot.FunctionalChaosAlgorithm(X, Y)
algo.run()
result = algo.getResult()

# %%
# Quick summary of sensitivity analysis
sensitivityAnalysis = ot.FunctionalChaosSobolIndices(result)
sensitivityAnalysis

# %%
# draw Sobol' indices
first_order = [sensitivityAnalysis.getSobolIndex(i) for i in range(dimension)]
total_order = [sensitivityAnalysis.getSobolTotalIndex(i) for i in range(dimension)]
graph = ot.SobolIndicesAlgorithm.DrawSobolIndices(input_names, first_order, total_order)
view = otv.View(graph)

# %%
# We saw that total order indices are close to first order,
# so the higher order indices must be all quite close to 0
for i in range(dimension):
    for j in range(i):
        sij = sensitivityAnalysis.getSobolIndex([i, j])
        print(f"{input_names[i]} & {input_names[j]}: {sij:.4f}")

# %%
# Display all figures
otv.View.ShowAll()



#! /usr/bin/env python

import openturns as ot

ot.TESTPREAMBLE()

distribution = ot.Multinomial(5, [0.25] * 3)
size = 10000
sample = distribution.getSample(size)
factory = ot.MultinomialFactory()
estimatedDistribution = factory.build(sample)
print("distribution=", repr(distribution))
print("Estimated distribution=", repr(estimatedDistribution))
estimatedDistribution = factory.build()
print("Default distribution=", estimatedDistribution)
estimatedMultinomial = factory.buildAsMultinomial(sample)
print("Multinomial          =", distribution)
print("Estimated multinomial=", estimatedMultinomial)
estimatedMultinomial = factory.buildAsMultinomial()
print("Default multinomial=", estimatedMultinomial)



"""
Calibration of the flooding model
=================================

The goal of this script is to produce the calibration.png image
that is printed in the documentation of the examples.
It is adapted from plot_calibration_flooding.py
"""

from openturns.usecases import flood_model
from matplotlib import pylab as plt
import openturns.viewer as viewer
import numpy as np
import openturns as ot

ot.ResourceMap.SetAsUnsignedInteger("Normal-SmallDimension", 1)
ot.Log.Show(ot.Log.NONE)

fm = flood_model.FloodModel()

Qobs = fm.data[:, 0]
Hobs = fm.data[:, 1]
nbobs = fm.data.getSize()


def functionFlooding(X):
    L = 5.0e3
    B = 300.0
    Q, K_s, Z_v, Z_m = X
    alpha = (Z_m - Z_v) / L
    if alpha < 0.0 or K_s <= 0.0:
        H = np.inf
    else:
        H = (Q / (K_s * B * np.sqrt(alpha))) ** (3.0 / 5.0)
    return [H]


g = ot.PythonFunction(4, 1, functionFlooding)
g = ot.MemoizeFunction(g)
g.setInputDescription(["Q ($m^3/s$)", "Ks ($m^{1/3}/s$)", "Zv (m)", "Zm (m)"])
g.setOutputDescription(["H (m)"])

KsInitial = 20.0
ZvInitial = 49.0
ZmInitial = 51.0
thetaPrior = [KsInitial, ZvInitial, ZmInitial]

calibratedIndices = [1, 2, 3]
mycf = ot.ParametricFunction(g, calibratedIndices, thetaPrior)

sigmaH = 0.5  # (m^2)

errorCovariance = ot.CovarianceMatrix(1)
errorCovariance[0, 0] = sigmaH**2

sigmaKs = 5.0
sigmaZv = 1.0
sigmaZm = 1.0

sigma = ot.CovarianceMatrix(3)
sigma[0, 0] = sigmaKs**2
sigma[1, 1] = sigmaZv**2
sigma[2, 2] = sigmaZm**2

defaultBootstrapSize = ot.ResourceMap.GetAsUnsignedInteger(
    "GaussianNonLinearCalibration-BootstrapSize"
)

ot.ResourceMap.SetAsUnsignedInteger("GaussianNonLinearCalibration-BootstrapSize", 0)

algo = ot.GaussianNonLinearCalibration(
    mycf, Qobs, Hobs, thetaPrior, sigma, errorCovariance
)
algo.run()
calibrationResult = algo.getResult()

grid = calibrationResult.drawParameterDistributions()
graph = grid.getGraph(0, 0)
graph.setLegends(grid.getGraph(0, 2).getLegends())
graph.setLegendPosition("topleft")
view = viewer.View(
    graph,
    figure_kw={"figsize": (4.0, 3.0)},
)
plt.subplots_adjust(bottom=0.2, left=0.2)
view.getFigure().savefig("calibration.png", bbox_inches="tight")
plt.show()

# Set back to default
ot.ResourceMap.SetAsUnsignedInteger(
    "GaussianNonLinearCalibration-BootstrapSize", defaultBootstrapSize
)



#! /usr/bin/env python

import openturns as ot

ot.TESTPREAMBLE()

# First, build two functions from R^3->R
inVar = ["x1", "x2", "x3"]
formula = ["x1^3 * sin(x2 + 2.5 * x3) - (x1 + x2)^2 / (1.0 + x3^2)"]
functions = []
functions.append(ot.SymbolicFunction(inVar, formula))
formula = ["exp(-x1 * x2 + x3) / cos(1.0 + x2 * x3 - x1)"]
functions.append(ot.SymbolicFunction(inVar, formula))
# Second, build the weights
coefficients = [0.3, 2.9]
# Third, build the function
myFunction = ot.LinearCombinationFunction(functions, coefficients)
inPoint = ot.Point([1.2, 2.3, 3.4])
print("myFunction=", myFunction)
print("Value at ", inPoint, "=", myFunction(inPoint))
print("Gradient at ", inPoint, "=", myFunction.gradient(inPoint))
print("Hessian at ", inPoint, "=", myFunction.hessian(inPoint))



#! /usr/bin/env python

import openturns as ot

ot.TESTPREAMBLE()

# Default dimension parameter to evaluate the model
defaultDimension = 1

# Amplitude values
amplitude = ot.Point(defaultDimension, 1.0)
# Scale values
scale = ot.Point(defaultDimension, 1.0)

# Spectral density
myModel = ot.CauchyModel(scale, amplitude)

points = 8
tMin = 0.0
tStep = 1.0 / (points - 1)

# RegularGrid --> Build list of frequencies using the RegularGrid
myTimeGrid = ot.RegularGrid(tMin, tStep, points)

mySpectralProcess0 = ot.SpectralGaussianProcess(myModel, myTimeGrid)

print("mySpectralProcess0 = ", mySpectralProcess0)
print("Realization = ", mySpectralProcess0.getRealization())

# Constructor using maximalFrequency value and size of discretization
maximalFrequency = 10.0
mySpectralProcess1 = ot.SpectralGaussianProcess(myModel, maximalFrequency, points)
tg = ot.RegularGrid(mySpectralProcess1.getTimeGrid())

print("mySpectralProcess1 = ", mySpectralProcess1)
print("Realization = ", mySpectralProcess1.getRealization())

# 3D model
highDimension = 3
amplitude = ot.Point(highDimension, 1.0)

# Second order model with parameters
mySpecModel = ot.CauchyModel(scale, amplitude)
print("mySpecModel = ", mySpecModel)

mySpectralProcess2 = ot.SpectralGaussianProcess(mySpecModel, myTimeGrid)
print("mySpectralProcess2 = ", mySpectralProcess2)
print("Realization = ", mySpectralProcess2.getRealization())
mySpectralProcess3 = ot.SpectralGaussianProcess(mySpecModel, maximalFrequency, points)
print("mySpectralProcess3 = ", mySpectralProcess3)
print("Realization = ", mySpectralProcess3.getRealization())



import openturns as ot
import openturns.experimental as otexp
import openturns.viewer as otv

from math import pi

ot.RandomGenerator.SetSeed(1)
log_UnscaledPDF = ot.SymbolicFunction("x", "log(cos(x)) + x")

eps = 1e-5
rangePDF = ot.Interval(-pi / 2.0 + eps, pi / 2.0 - eps)
ratioAlgo = otexp.RatioOfUniforms(log_UnscaledPDF, rangePDF)

unscaled_PDF = ot.ComposedFunction(ot.SymbolicFunction("x", "exp(x)"), log_UnscaledPDF)
norm_fact = ot.GaussLegendre().integrate(unscaled_PDF, rangePDF)[0]
exact_PDF = ot.LinearCombinationFunction([unscaled_PDF], [1.0 / norm_fact])

size = 100000
sample = ratioAlgo.getSample(size)

ks_algo = ot.KernelSmoothing()
ks_algo.setBoundaryCorrection(True)
ks_algo.setLowerBound(-pi / 2.0)
ks_algo.setUpperBound(pi / 2.0)
ks_pdf = ks_algo.build(sample)

g = ks_pdf.drawPDF()
draw = exact_PDF.draw(-pi / 2.0, pi / 2.0, 1001).getDrawable(0)
draw.setLineWidth(2)
g.add(draw)
g.setLegends(["Ratio of Uniforms", "exact pdf"])
g.setLegendPosition("topleft")
g.setXTitle("x")

view = otv.View(g)



#! /usr/bin/env python

import openturns as ot
import openturns.experimental as otexp
import openturns.testing as ott

ot.TESTPREAMBLE()

objective = ot.SymbolicFunction(["x0", "x1"], ["1.1*x0 + x1"])
ineq = ot.SymbolicFunction(["x0", "x1"], ["x0 + 2*x1 - 5", "3*x0 + 2*x1 - 6"])
eq = ot.SymbolicFunction(["x0", "x1"], ["7*x0 + 9*x1 - 50"])
problem = ot.OptimizationProblem(objective)
problem.setInequalityConstraint(ineq)
problem.setEqualityConstraint(eq)
bounds = ot.Interval([0.0, 1.0], [4.0, 1e30])
problem.setBounds(bounds)
problem.setVariablesType(
    [
        ot.OptimizationProblemImplementation.CONTINUOUS,
        ot.OptimizationProblemImplementation.INTEGER,
    ]
)
location = [0.0] * 2
linearProblem = otexp.LinearProblem.Linearize(problem, location)
print(linearProblem)
cost = linearProblem.getLinearCost()
A = linearProblem.getLinearConstraintCoefficients()
LU = linearProblem.getLinearConstraintBounds()
assert cost == [1.1, 1.0]
ott.assert_almost_equal(A, ot.Matrix([[1, 2], [3, 2], [7, 9]]))
assert LU.getLowerBound() == [5.0, 6.0, 50.0]
assert LU.getUpperBound()[2] == 50.0
assert problem.getVariablesType()[1] == ot.OptimizationProblemImplementation.INTEGER
assert problem.getBounds().getLowerBound() == [0.0, 1.0]



"""
Fit an extreme value distribution
=================================
"""

# %%
import openturns as ot
import openturns.viewer as otv

# %%
# The generalized extreme value distribution (GEV)
# ------------------------------------------------

# %%
# The :class:`~openturns.GeneralizedExtremeValue` distribution is a family of continuous probability distributions
# that combine the :class:`~openturns.Gumbel`, :class:`~openturns.Frechet` and :class:`~openturns.WeibullMax` distributions, all extreme value distributions.
#
# In this example we use the associated :class:`~openturns.GeneralizedExtremeValueFactory` to fit sample with extreme values.
# This factory returns the best model among the Frechet, Gumbel and Weibull estimates according to the BIC criterion.
#


# %%
# We draw a sample from a Gumbel of parameters :math:`\beta = 1.0` and :math:`\gamma = 3.0`
# and another one from a Frechet with parameters :math:`\beta=1.0`, :math:`\alpha = 1.0` and :math:`\gamma = 0.0`.
# We consider both samples as a single sample from an unknown extreme distribution to be fitted.
#

# %%
# The distributions used:
myGumbel = ot.Gumbel(1.0, 3.0)
myFrechet = ot.Frechet(1.0, 1.0, 0.0)


# %%
# We build our experiment sample of size 2000.
sample = ot.Sample()
sampleFrechet = myFrechet.getSample(1000)
sampleGumbel = myGumbel.getSample(1000)
sample.add(sampleFrechet)
sample.add(sampleGumbel)


# %%
# We fit the sample thanks to the :class:`~openturns.GeneralizedExtremeValueFactory`:
myDistribution = ot.GeneralizedExtremeValueFactory().buildAsGeneralizedExtremeValue(
    sample
)

# %%
# We can display the parameters of the fitted distribution `myDistribution`:
print(myDistribution)

# %%
# We can also get the actual distribution (Weibull, Frechet or Gumbel) with the `getActualDistribution` method:
print(myDistribution.getActualDistribution())

# %%
# The given sample is then best described by a Frechet distribution.

# %%
# We draw the fitted distribution and a histogram of the data.
graph = myDistribution.drawPDF()
graph.add(ot.HistogramFactory().build(sample).drawPDF())
graph.setLegends(["GEV fitting", "histogram"])
graph.setLegendPosition("upper right")

view = otv.View(graph)
axes = view.getAxes()
_ = axes[0].set_xlim(-20.0, 20.0)


# %%
# We compare different fitting strategies for this sample:
#
# - use the histogram from the data (red)
# - the GEV fitted distribution (black)
# - the pure Frechet fitted distribution (green)
# - the pure Gumbel fitted distribution (blue)
# - the pure WeibullMax fitted distribution (dashed orange)
#
graph = myDistribution.drawPDF()
graph.add(ot.HistogramFactory().build(sample).drawPDF())

distFrechet = ot.FrechetFactory().buildAsFrechet(sample)
graph.add(distFrechet.drawPDF())

distGumbel = ot.GumbelFactory().buildAsGumbel(sample)
graph.add(distGumbel.drawPDF())

# We change the line style of the WeibullMax.
distWeibullMax = ot.WeibullMaxFactory().buildAsWeibullMax(sample)
curveWeibullMax = distWeibullMax.drawPDF().getDrawable(0)
curveWeibullMax.setLineStyle("dashed")
graph.add(curveWeibullMax)

graph.setLegends(
    [
        "GEV fitting",
        "histogram",
        "Frechet fitting",
        "Gumbel fitting",
        "WeibullMax fitting",
    ]
)
graph.setLegendPosition("upper right")
view = otv.View(graph)
axes = view.getAxes()  # axes is a matplotlib object
_ = axes[0].set_xlim(-20.0, 20.0)


# %%
# As returned by the `getActualDistribution` method the GEV distribution is a Frechet.
# The :class:`~openturns.GeneralizedExtremeValueFactory` class is a convenient class to fit extreme valued samples
# without an a priori knowledge of the underlying (at least the closest) extreme distribution.


# %%
# The Generalized Pareto Distribution (GPD)
# -----------------------------------------
#
# In this paragraph we fit a :class:`~openturns.GeneralizedPareto` distribution.
# Various estimators are provided by the GPD factory. Please refer to the :class:`~openturns.GeneralizedParetoFactory` class documentation for more information.
# The selection is based on the sample size and compared to the `GeneralizedParetoFactory-SmallSize` key of the :class:`~openturns.ResourceMap`:
#

# %%
print(ot.ResourceMap.GetAsUnsignedInteger("GeneralizedParetoFactory-SmallSize"))


# %%
# The small sample case
# ^^^^^^^^^^^^^^^^^^^^^
#
# In this case the default estimator is based on a probability weighted method of moments, with a fallback on the exponential regression method.

# %%
myDist = ot.GeneralizedPareto(1.0, 0.0, 0.0)
N = 17
sample = myDist.getSample(N)

# %%
# We build our experiment sample of size `N`.
myFittedDist = ot.GeneralizedParetoFactory().buildAsGeneralizedPareto(sample)
print(myFittedDist)

# %%
# We draw the fitted distribution as well as an histogram to visualize the fit:
graph = myFittedDist.drawPDF()
graph.add(ot.HistogramFactory().build(sample).drawPDF())
graph.setTitle("Generalized Pareto distribution fitting on a sample")
graph.setLegends(["GPD fitting", "histogram"])
graph.setLegendPosition("upper right")

view = otv.View(graph)
axes = view.getAxes()
_ = axes[0].set_xlim(-1.0, 10.0)


# %%
# Large sample
# ^^^^^^^^^^^^
#
# For a larger sample the estimator is based on an exponential regression method
# with a fallback on the probability weighted moments estimator.


# %%
N = 1000
sample = myDist.getSample(N)

# %%
# We build our experiment sample of size `N`.
myFittedDist = ot.GeneralizedParetoFactory().buildAsGeneralizedPareto(sample)
print(myFittedDist)

# %%
# We draw the fitted distribution as well as a histogram to visualize the fit:
graph = myFittedDist.drawPDF()
graph.add(ot.HistogramFactory().build(sample).drawPDF())
graph.setTitle("Generalized Pareto distribution fitting on a sample")
graph.setLegends(["GPD fitting", "histogram"])
graph.setLegendPosition("upper right")

view = otv.View(graph)
axes = view.getAxes()
_ = axes[0].set_xlim(-1.0, 10.0)

# %%
# Display all graphs
otv.View.ShowAll()



import openturns as ot

data_ref = ot.Sample.ImportFromCSVFile("besselI1_maple.csv")
print(data_ref, data_ref.getSize())
errMax = 0
err1 = 100
err2 = 100
for j in range(data_ref.getSize()):
    x = data_ref[j, 0]
    ref = data_ref[j, 1]
    val1 = ot.SpecFunc.BesselI1(x)
    val2 = ot.SpecFunc.BesselI2(x)
    err1 = abs(ref - val1)
    err2 = abs(ref - val2)
    if ref != 0.0:
        err1 /= abs(ref)
        err2 /= abs(ref)
    if err1 < err2:
        print("111111111 x=", x, "error 1=", err1, "error 2=", err2)
    else:
        print("222222222 x=", x, "error 1=", err1, "error 2=", err2)



#! /usr/bin/env python

import openturns as ot
from matplotlib.backends.backend_pdf import PdfPages
from openturns.viewer import View
import time

ot.RandomGenerator.SetSeed(0)
ot.Log.Show(ot.Log.INFO)

# Bounds are [0,1]^dimension
dimension = 50

nSimu = 10000
c2 = ot.SpaceFillingC2()
pp = PdfPages("large_mc_OTLHS.pdf")

# Size of sample
size = 100
# Factory: lhs generates
lhsDesign = ot.LHSExperiment(
    ot.JointDistribution([ot.Uniform(0.0, 1.0)] * dimension), size
)
lhsDesign.setAlwaysShuffle(True)  # randomized

mc = ot.MonteCarloLHS(lhsDesign, nSimu, c2)
tic = time.time()
design = mc.generate()
result = mc.getResult()
toc = time.time()
print("cpu time=%f" % (toc - tic))
print("dimension=%d, size=%d,mc=%s" % (dimension, size, mc))
print(
    "optimal value="
    + str(result.getOptimalValue())
    + " c2="
    + str(result.getC2())
    + " phiP="
    + str(result.getPhiP())
    + " minDist="
    + str(result.getMinDist())
)
# plot criterion
crit = result.drawHistoryCriterion()
# in pdf
pp.savefig(View(crit, plot_kwargs={"color": "blue"}).getFigure())

minDist = ot.SpaceFillingMinDist()
# Factory: lhs generates
lhsDesign = ot.LHSExperiment(
    ot.JointDistribution([ot.Uniform(0.0, 1.0)] * dimension), size
)
lhsDesign.setAlwaysShuffle(True)  # randomized

mc = ot.MonteCarloLHS(lhsDesign, nSimu, minDist)
tic = time.time()
result = mc.generate()
toc = time.time()
print("cpu time=%f" % (toc - tic))
print("dimension=%d, size=%d,mc=%s" % (dimension, size, mc))
print(
    "optimal value="
    + str(result.getOptimalValue())
    + " c2="
    + str(result.getC2())
    + " phiP="
    + str(result.getPhiP())
    + " minDist="
    + str(result.getMinDist())
)
# plot criterion
crit = result.drawHistoryCriterion()
pp.savefig(View(crit, plot_kwargs={"color": "blue"}).getFigure())

pp.close()



#!/usr/bin/env python

import openturns as ot
import openturns.testing as ott

ot.TESTPREAMBLE()
# ot.Log.Show(ot.Log.ALL)
dim = 7

# root cause
mean = [200.0] * dim
mean[-1] = 60
mean[-2] = 60
sigma = [30.0] * dim
sigma[-1] = 15.0
R = ot.CorrelationMatrix(dim)
for i in range(5):
    for j in range(i):
        R[i, j] = 0.5
dist = ot.Normal(mean, sigma, R)

# leaf events
X = ot.RandomVector(dist)
inputs = ["M1", "M2", "M3", "M4", "M5", "H", "V"]
e0 = ot.ThresholdEvent(
    ot.CompositeRandomVector(ot.SymbolicFunction(inputs, ["M1+M2+M4+M5-5*H"]), X),
    ot.Less(),
    0.0,
)
e1 = ot.ThresholdEvent(
    ot.CompositeRandomVector(ot.SymbolicFunction(inputs, ["M2+2*M3+M4-5*V"]), X),
    ot.Less(),
    0.0,
)
e2 = ot.ThresholdEvent(
    ot.CompositeRandomVector(
        ot.SymbolicFunction(inputs, ["M1+2*M3+2*M4+M5-5*H-5*V"]), X
    ),
    ot.Less(),
    0.0,
)
e3 = ot.ThresholdEvent(
    ot.CompositeRandomVector(ot.SymbolicFunction(inputs, ["-(M1+M2+M4+M5-5*80.0)"]), X),
    ot.Less(),
    0.0,
)
e4 = ot.ThresholdEvent(
    ot.CompositeRandomVector(ot.SymbolicFunction(inputs, ["-(M2+2*M3+M4-5*70.0)"]), X),
    ot.Less(),
    0.0,
)
e5 = ot.ThresholdEvent(
    ot.CompositeRandomVector(
        ot.SymbolicFunction(inputs, ["-(M1+2*M3+2*M4+M5-5*80.0-5*70.0)"]), X
    ),
    ot.Less(),
    0.0,
)

# solver
solver = ot.AbdoRackwitz()
solver.setStartingPoint(mean)
solver.setMaximumIterationNumber(1000)
solver.setMaximumAbsoluteError(1.0e-3)
solver.setMaximumRelativeError(1.0e-3)
solver.setMaximumResidualError(1.0e-3)
solver.setMaximumConstraintError(1.0e-3)

for event in [e0, e3, e4, e5, e1, e2]:
    algo = ot.FORM(solver, event)
    algo.run()
    result = algo.getResult()
    print("beta = %.6g" % result.getGeneralisedReliabilityIndex())

# system event in DNF form (union of intersections)
event = ot.UnionEvent(
    [
        ot.IntersectionEvent([e0, e3, e4, e5]),
        ot.IntersectionEvent([e1, e3, e4, e5]),
        # ot.IntersectionEvent([e2, e3, e4, e5]),
    ]
)

# sampling test
pf_sim = event.getSample(10000).computeMean()[0]
print("pf_sim = %.6g" % pf_sim)
ott.assert_almost_equal(pf_sim, 0.0023, 1e-3, 1e-3)

# system FORM
algo = ot.SystemFORM(solver, event)
algo.run()
result = algo.getResult()
pf_sysform = result.getEventProbability()
print("pf_sysform = %.6g" % pf_sysform)
ott.assert_almost_equal(pf_sysform, 0.00234983, 1e-4, 1e-4)

for form_result in result.getFORMResultCollection():
    print(" beta=", form_result.getGeneralisedReliabilityIndex())



#! /usr/bin/env python

import openturns as ot

ot.TESTPREAMBLE()

# Beta related functions
# pBeta
p1Min = 0.2
p1Max = 5.0
n1 = 5
p2Min = 0.2
p2Max = 5.0
n2 = 5
xMin = 0.1
xMax = 0.9
nX = 5
for i1 in range(n1):
    p1 = p1Min + (p1Max - p1Min) * i1 / (n1 - 1)
    for i2 in range(n2):
        p2 = p2Min + (p2Max - p2Min) * i2 / (n2 - 1)
        for iX in range(nX):
            x = xMin + (xMax - xMin) * iX / (nX - 1)
            print(
                "pBeta(",
                p1,
                ", ",
                p2,
                ",  %.12g" % x,
                ")=%.6g" % ot.DistFunc.pBeta(p1, p2, x),
                ", complementary=%.6g" % ot.DistFunc.pBeta(p1, p2, x, True),
            )
# qBeta
p1Min = 0.2
p1Max = 5.0
n1 = 5
p2Min = 0.2
p2Max = 5.0
n2 = 5
qMin = 0.1
qMax = 0.9
nQ = 5
for i1 in range(n1):
    p1 = p1Min + (p1Max - p1Min) * i1 / (n1 - 1)
    for i2 in range(n2):
        p2 = p2Min + (p2Max - p2Min) * i2 / (n2 - 1)
        for iQ in range(nQ):
            q = qMin + (qMax - qMin) * iQ / (nQ - 1)
            print(
                "qBeta(",
                p1,
                ", ",
                p2,
                ",  %.12g" % q,
                ")=%.6g" % ot.DistFunc.qBeta(p1, p2, q),
                ", complementary=%.6g" % ot.DistFunc.qBeta(p1, p2, q, True),
            )
# rBeta
p1Min = 0.2
p1Max = 5.0
n1 = 5
p2Min = 0.2
p2Max = 5.0
n2 = 5
nR = 5
for i1 in range(n1):
    p1 = p1Min + (p1Max - p1Min) * i1 / (n1 - 1)
    for i2 in range(n2):
        p2 = p2Min + (p2Max - p2Min) * i2 / (n2 - 1)
        for iR in range(nR):
            print("rBeta(", p1, ", ", p2, ")=%.6g" % ot.DistFunc.rBeta(p1, p2))



#! /usr/bin/env python

import openturns as ot

# Polynomial factories
factoryCollection = [ot.MonomialFunctionFactory()] * 3

dim = len(factoryCollection)
factory = ot.TensorizedUniVariateFunctionFactory(factoryCollection)
print("factory=", factory)
x = [0.5, 1.0, 1.5]
for i in range(10):
    f = factory.build(i)
    print("i=", i, "f(X)=", f(x))



#! /usr/bin/env python

import openturns as ot
import openturns.testing as ott

ot.TESTPREAMBLE()

# Instantiate one distribution object
dim = 3
copula = ot.IndependentCopula(dim)
print("Copula =", repr(copula))
print("Copula =", copula)

# Is this copula an elliptical distribution?
print("Elliptical distribution = ", copula.isElliptical())

# Is this copula continuous ?
print("Continuous copula = ", copula.isContinuous())

# Is this copula elliptical ?
print("Elliptical copula = ", copula.hasEllipticalCopula())

# Is this copula independent ?
print("hasIndependentCopula = ", copula.hasIndependentCopula())

# Test for realization of copula
oneRealization = copula.getRealization()
print("oneRealization=", repr(oneRealization))

# Define a point
point = ot.Point(copula.getDimension(), 0.6)
print("Point= ", repr(point))

# derivative of PDF with regards its arguments
DDF = copula.computeDDF(point)
print("ddf     =", repr(DDF))

# PDF value
PDF = copula.computePDF(point)
print("pdf     =%.6f" % PDF)

# CDF value
CDF = copula.computeCDF(point)
print("cdf=%.6f" % CDF)

# derivative of the PDF with regards the parameters of the distribution
PDFgr = copula.computePDFGradient(point)
print("pdf gradient     =", repr(PDFgr))

# derivative of the CDF with regards the parameters of the distribution
CDFgr = copula.computeCDFGradient(point)
print("cdf gradient     =", repr(CDFgr))

# quantile
quantile = copula.computeQuantile(0.95)
print("quantile=", repr(quantile))
print("cdf(quantile)=%.6f" % copula.computeCDF(quantile))
# Get 95% survival function
inverseSurvival = ot.Point(copula.computeInverseSurvivalFunction(0.95))
print("InverseSurvival=", repr(inverseSurvival))
print(
    "Survival(inverseSurvival)=%.6f" % copula.computeSurvivalFunction(inverseSurvival)
)
print("entropy=%.6f" % copula.computeEntropy())
# Confidence regions
interval, threshold = copula.computeMinimumVolumeIntervalWithMarginalProbability(0.95)
print("Minimum volume interval=", interval)
print("threshold=", ot.Point(1, threshold))
levelSet, beta = copula.computeMinimumVolumeLevelSetWithThreshold(0.95)
print("Minimum volume level set=", levelSet)
print("beta=", ot.Point(1, beta))
interval, beta = copula.computeBilateralConfidenceIntervalWithMarginalProbability(0.95)
print("Bilateral confidence interval=", interval)
print("beta=", ot.Point(1, beta))
interval, beta = copula.computeUnilateralConfidenceIntervalWithMarginalProbability(
    0.95, False
)
print("Unilateral confidence interval (lower tail)=", interval)
print("beta=", ot.Point(1, beta))
interval, beta = copula.computeUnilateralConfidenceIntervalWithMarginalProbability(
    0.95, True
)
print("Unilateral confidence interval (upper tail)=", interval)
print("beta=", ot.Point(1, beta))

# mean
mean = copula.getMean()
print("mean=", repr(mean))

# covariance
covariance = copula.getCovariance()
print("covariance=", repr(covariance))

# parameters of the distribution
parameters = copula.getParametersCollection()
print("parameters=", repr(parameters))

# Specific to this copula

# Extract the marginals
for i in range(dim):
    margin = copula.getMarginal(i)
    print("margin=", repr(margin))
    print("margin PDF=%.6f" % margin.computePDF(ot.Point(1, 0.25)))
    print("margin CDF=%.6f" % margin.computeCDF(ot.Point(1, 0.25)))
    print("margin quantile=", repr(margin.computeQuantile(0.95)))
    print("margin realization=", repr(margin.getRealization()))

# Extract a 2-D marginal
indices = [1, 0]
print("indices=", repr(indices))
margins = copula.getMarginal(indices)
print("margins=", repr(margins))
print("margins PDF=%.6f" % margins.computePDF(ot.Point(2, 0.25)))
print("margins CDF=%.6f" % margins.computeCDF(ot.Point(2, 0.25)))
quantile = margins.computeQuantile(0.95)
print("margins quantile=", repr(quantile))
print("margins CDF(quantile)=%.6f" % margins.computeCDF(quantile))
print("margins realization=", repr(margins.getRealization()))

ot.Log.Show(ot.Log.TRACE)
validation = ott.DistributionValidation(copula)
validation.run()



#! /usr/bin/env python

import openturns as ot

ot.TESTPREAMBLE()

# We create a numerical math function
myFunction = ot.SymbolicFunction(("E", "F", "L", "I"), ("-F*L^3/(3.*E*I)",))
myFunction = ot.MemoizeFunction(myFunction)
dim = myFunction.getInputDimension()

# We create a normal distribution point of dimension 1
mean = ot.Point(dim, 0.0)
# E
mean[0] = 50.0
# F
mean[1] = 1.0
# L
mean[2] = 10.0
# I
mean[3] = 5.0
sigma = ot.Point(dim, 1.0)
R = ot.IdentityMatrix(dim)
myDistribution = ot.Normal(mean, sigma, R)

# We create a 'usual' RandomVector from the Distribution
vect = ot.RandomVector(myDistribution)

# We create a composite random vector
output = ot.CompositeRandomVector(myFunction, vect)

# We create an Event from this RandomVector
myEvent = ot.ThresholdEvent(output, ot.Less(), -3.0)

# We create a Monte Carlo algorithm
experiment = ot.MonteCarloExperiment()
myAlgo = ot.ProbabilitySimulationAlgorithm(myEvent, experiment)
myAlgo.setMaximumOuterSampling(500)
myAlgo.setBlockSize(10)
myAlgo.setMaximumCoefficientOfVariation(0.05)

print("MonteCarlo=", myAlgo)

# Perform the simulation
myAlgo.run()

# Stream out the result
result = myAlgo.getResult()
print("MonteCarlo result=", result)

# Compute sensitivity information
print("mean point in event domain=", result.getMeanPointInEventDomain())
print("importance factors=", result.getImportanceFactors())



#! /usr/bin/env python

import openturns as ot
import openturns.testing as ott

ot.TESTPREAMBLE()

# Instantiate one distribution object
distribution = ot.KPermutationsDistribution(5, 12)
print("Distribution ", distribution)

# Is this distribution elliptical ?
print("Elliptical = ", distribution.isElliptical())

# Is this distribution continuous ?
print("Continuous = ", distribution.isContinuous())

# Test for realization of distribution
oneRealization = distribution.getRealization()
print("oneRealization=", oneRealization)

# Define a point
point = ot.Point(distribution.getDimension(), 4.0)
print("Point= ", point)

# Show PDF and CDF of point
LPDF = distribution.computeLogPDF(point)
assert LPDF == ot.SpecFunc.LowestScalar
PDF = distribution.computePDF(point)
print("pdf     =", PDF)
CDF = distribution.computeCDF(point)
print("cdf=%.6f" % CDF)
CCDF = distribution.computeComplementaryCDF(point)
print("ccdf=%.6f" % CCDF)
quantile = distribution.computeQuantile(0.95)
print("quantile=", quantile)
print("cdf(quantile)=", distribution.computeCDF(quantile))
print("entropy=%.6f" % distribution.computeEntropy())
mean = distribution.getMean()
print("mean=", mean)
covariance = distribution.getCovariance()
print("covariance=", covariance)
parameters = distribution.getParametersCollection()
print("parameters=", parameters)

ot.Log.Show(ot.Log.TRACE)
validation = ott.DistributionValidation(distribution)
validation.skipMoments()  # slow
validation.skipCorrelation()  # slow
validation.run()



"""
Quick start guide to optimization
=================================
"""

# %%
#
# In this example, we perform the optimization of the Rosenbrock test function.
#
# Let :math:`a, b\in\mathbb{R}` be parameters. The Rosenbrock function is defined by
#
# .. math::
#    f(x_1, x_2) = (a-x_1)^2 + b(x_2 - x_1^2)^2
#
#
# for any :math:`\mathbf{x}\in\mathbb{R}^2`.
# This function is often used with :math:`a=1` and :math:`b=100`. In this case, the function has a single global minimum at:
#
# .. math::
#    \mathbf{x}^\star = (1, 1)^T.
#
#
# This function has a nonlinear least squares structure.
#
# *References*
#
# * Rosenbrock, H.H. (1960). "An automatic method for finding the greatest or least value of a function". The Computer Journal. 3 (3): 175–184.

# %%
# Definition of the function
# --------------------------

# %%
import openturns as ot
import openturns.viewer as otv


# %%
rosenbrock = ot.SymbolicFunction(["x1", "x2"], ["(1-x1)^2+100*(x2-x1^2)^2"])

# %%
x0 = [-1.0, 1.0]

# %%
xexact = ot.Point([1.0, 1.0])

# %%
lowerbound = [-2.0, -2.0]
upperbound = [2.0, 2.0]

# %%
# Plot the iso-values of the objective function
# ---------------------------------------------

# %%
rosenbrock = ot.MemoizeFunction(rosenbrock)

# %%
graph = rosenbrock.draw(lowerbound, upperbound, [100] * 2)
graph.setTitle("Rosenbrock function")
view = otv.View(graph)

# %%
# We see that the minimum is on the top right of the picture and the starting point is on the top left of the picture.
# Since the function has a long valley following the curve :math:`x_2 - x^2=0`, the algorithm generally have to follow the bottom of the valley.

# %%
# Create and solve the optimization problem
# -----------------------------------------

# %%
problem = ot.OptimizationProblem(rosenbrock)

# %%
algo = ot.Cobyla(problem)
algo.setMaximumRelativeError(1.0e-1)  # on x
algo.setMaximumCallsNumber(50000)
algo.setStartingPoint(x0)
algo.run()

# %%
result = algo.getResult()

# %%
xoptim = result.getOptimalPoint()
xoptim

# %%
delta = xexact - xoptim
absoluteError = delta.norm()
absoluteError

# %%
# We see that the algorithm found an accurate approximation of the solution.

# %%
result.getOptimalValue()  # f(x*)

# %%
result.getCallsNumber()

# %%
graph = rosenbrock.draw(lowerbound, upperbound, [100] * 2)
cloud = ot.Cloud(ot.Sample([x0, xoptim]))
cloud.setColor("black")
cloud.setPointStyle("bullet")
graph.add(cloud)
graph.setTitle("Rosenbrock function")
view = otv.View(graph)

# %%
# We see that the algorithm had to start from the top left of the banana and go to the top right.

# %%
graph = result.drawOptimalValueHistory()
view = otv.View(graph)

# %%
# The function value history make the path of the algorithm clear.
# In the first step, the algorithm went in the valley, which made the function value decrease rapidly.
# Once there, the algorithm had to follow the bottom of the valley so that the function decreased but slowly.
# In the final steps, the algorithm found the neighbourhood of the minimum so that the local convergence could take place.

# %%
# In order to see where the function was evaluated, we use the `getInputSample` method.

# %%
inputSample = result.getInputSample()

# %%
graph = rosenbrock.draw(lowerbound, upperbound, [100] * 2)
graph.setTitle("Rosenbrock function solved with Cobyla")
cloud = ot.Cloud(inputSample)
graph.add(cloud)
view = otv.View(graph)

# %%
# We see that the algorithm made lots of evaluations in the bottom of the valley before getting in the neighbourhood of the minimum.

# %%
# Solving the problem with NLopt
# ------------------------------
#
# We see that the `Cobyla` algorithm required lots of function evaluations.
# This is why we now use the `NLopt` class with the LBFGS algorithm.
# However, the algorithm may use input points which are far away from the input domain we used so far.
# This is why we had bounds to the problem, so that the algorithm never goes to far away from the valley.

# %%
bounds = ot.Interval(lowerbound, upperbound)

# %%
problem = ot.OptimizationProblem(rosenbrock)
problem.setBounds(bounds)

# %%
algo = ot.NLopt(problem, "LD_LBFGS")
algo.setStartingPoint(x0)
algo.run()

# %%
result = algo.getResult()

# %%
xoptim = result.getOptimalPoint()
xoptim

# %%
delta = xexact - xoptim
absoluteError = delta.norm()
absoluteError

# %%
# We see that the algorithm found an extremely accurate approximation of the solution.

# %%
result.getOptimalValue()  # f(x*)

# %%
result.getCallsNumber()

# %%
# This number of iterations is much less than the previous experiment.

# %%
inputSample = result.getInputSample()

# %%
graph = rosenbrock.draw(lowerbound, upperbound, [100] * 2)
graph.setTitle("Rosenbrock function solved with NLopt/LD_LBFGS")
cloud = ot.Cloud(inputSample)
graph.add(cloud)
view = otv.View(graph)

# %%
# Display all figures
otv.View.ShowAll()



#! /usr/bin/env python

import openturns as ot
from math import pi

ot.TESTPREAMBLE()

# Log.Show( Log.Flags() | Log.INFO )

# Problem parameters
dimension = 3
a = 7.0
b = 0.1
# Reference analytical values
meanTh = a / 2
covTh = (b**2 * pi**8) / 18.0 + (b * pi**4) / 5.0 + (a**2) / 8.0 + 1.0 / 2.0
sob_1 = [
    (b * pi**4 / 5.0 + b**2 * pi**8 / 50.0 + 1.0 / 2.0) / covTh,
    (a**2 / 8.0) / covTh,
    0.0,
]
sob_2 = [0.0, (b**2 * pi**8 / 18.0 - b**2 * pi**8 / 50.0) / covTh, 0.0]
sob_3 = [0.0]
sob_T1 = [
    sob_1[0] + sob_2[0] + sob_2[1] + sob_3[0],
    sob_1[1] + sob_2[0] + sob_2[2] + sob_3[0],
    sob_1[2] + sob_2[1] + sob_2[2] + sob_3[0],
]
sob_T2 = [
    sob_2[0] + sob_2[1] + sob_3[0],
    sob_2[0] + sob_2[2] + sob_3[0],
    sob_2[1] + sob_2[2] + sob_3[0],
]
sob_T3 = [sob_3[0]]
# Create the Ishigami function
inputVariables = ["xi1", "xi2", "xi3"]
formula = [
    "sin(xi1) + (" + str(a) + ") * (sin(xi2)) ^ 2 + (" + str(b) + ") * xi3^4 * sin(xi1)"
]
model = ot.SymbolicFunction(inputVariables, formula)

# Create the input distribution
distribution = ot.JointDistribution([ot.Uniform(-pi, pi)] * dimension)

# Create the orthogonal basis
enumerateFunction = ot.LinearEnumerateFunction(dimension)
productBasis = ot.OrthogonalProductPolynomialFactory(
    [ot.LegendreFactory()] * dimension, enumerateFunction
)

# design experiment
samplingSize = 75

# build basis
degree = 10
basisSize = enumerateFunction.getStrataCumulatedCardinal(degree)

# run algorithm
lar = ot.LARS()

listFittingAlgorithm = list()
listFittingAlgorithm.append(ot.KFold())
listFittingAlgorithm.append(ot.CorrectedLeaveOneOut())

for fittingAlgorithmIndex in range(len(listFittingAlgorithm)):
    fittingAlgorithm = listFittingAlgorithm[fittingAlgorithmIndex]
    adaptiveStrategy = ot.FixedStrategy(productBasis, basisSize)
    projectionStrategy = ot.LeastSquaresStrategy(
        ot.LeastSquaresMetaModelSelectionFactory(ot.LARS(), fittingAlgorithm)
    )
    experiment = ot.LowDiscrepancyExperiment(
        ot.SobolSequence(), distribution, samplingSize
    )
    ot.RandomGenerator.SetSeed(0)
    X = experiment.generate()
    Y = model(X)
    algo = ot.FunctionalChaosAlgorithm(
        X, Y, distribution, adaptiveStrategy, projectionStrategy
    )
    algo.run()

    result = algo.getResult()
    print("coeffs = ", result.getCoefficients())
    print("isLeastSquares= ", result.isLeastSquares())
    assert result.isLeastSquares()
    print("involvesModelSelection= ", result.involvesModelSelection())
    assert result.involvesModelSelection()



#! /usr/bin/env python

import openturns as ot
import openturns.testing as ott

ot.TESTPREAMBLE()

# need a proper LS solver
if not ot.PlatformInfo.HasFeature("cminpack") or not ot.PlatformInfo.HasFeature(
    "ceres"
):
    exit(0)

size = 10000

# gumbel
distribution = ot.Gumbel(1.5, -0.5)
print("distribution=", distribution)
sample = distribution.getSample(size)
factory = ot.QuantileMatchingFactory(ot.Gumbel(), [0.01, 0.99])
inf_dist = factory.build(sample)
print("estimated distribution=", inf_dist)
ott.assert_almost_equal(
    distribution.getParameter(), inf_dist.getParameter(), 1e-2, 1e-2
)

# set (a,b) out of (r, t, a, b)
distribution = ot.Beta(2.3, 2.2, -1.0, 1.0)
p_ref = distribution.getParameter()
print("distribution=", distribution)
sample = distribution.getSample(size)
factory = ot.QuantileMatchingFactory(ot.Beta(), [0.01, 0.99])
factory.setKnownParameter([2, 3], [-1.0, 1.0])
inf_dist = factory.build(sample)
print("estimated distribution=", inf_dist)
ott.assert_almost_equal(inf_dist.getParameter(), p_ref, 1e-2, 1e-2)

# full probas
probabilities = [0.01, 1 / 3, 2 / 3, 0.99]
factory = ot.QuantileMatchingFactory(ot.Beta(), probabilities)
quantiles = [distribution.computeQuantile(pi)[0] for pi in probabilities]
inf_dist = factory.buildFromQuantiles(quantiles)
print("estimated distribution (quantiles)=", inf_dist)
ott.assert_almost_equal(inf_dist.getParameter(), p_ref, 1e-2, 1e-2)

# from quantiles, known parameters
factory = ot.QuantileMatchingFactory(ot.Beta(), [0.01, 0.99])
factory.setKnownParameter([2, 3], [-1.0, 1.0])
quantiles = [-0.847, 0.867]
inf_dist = factory.buildFromQuantiles(quantiles)
print("estimated distribution (quantiles)=", inf_dist)
probabilities = factory.getProbabilities()
ott.assert_almost_equal(inf_dist.getParameter(), p_ref, 1e-2, 1e-2)

# ill defined case
probabilities = [0.05, 0.25, 0.75, 0.95]
factory = ot.QuantileMatchingFactory(ot.Beta(), probabilities)
quantiles = [-4.0, 1.0, 2.0, 10.0]
ok = False
try:
    match_dist = factory.buildFromQuantiles(quantiles)
except Exception:
    ok = True
assert ok, "should raise"

# Normal, optimal far away
distribution = ot.Normal(17.0, 1.0)
print("distribution=", distribution)
sample = distribution.getSample(size)
factory = ot.QuantileMatchingFactory(ot.Normal(), [0.01, 0.99])
inf_dist = factory.build(sample)
ott.assert_almost_equal(
    inf_dist.getParameter(), distribution.getParameter(), 1e-2, 1e-2
)



"""
Random generator parametrization
================================
"""

# %%
# The seed of the pseudo random generator is initialized to 0 when the module is imported.
#
# It means the same script will yield the same results across several replications.
#
# It is also possible to initialize the random generator differently:
#
# - by setting a seed
# - by setting the complete generator state

# %%
import openturns as ot
import os
import time


# %%
# Example 0: using the time in milliseconds
# -----------------------------------------
ot.RandomGenerator.SetSeed(int(1000 * time.time() % 1e9))

# %%
# Example 1: using the python process id
# --------------------------------------
ot.RandomGenerator.SetSeed(os.getpid())

# %%
# Example 2: using a fixed seed
# -----------------------------
ot.RandomGenerator.SetSeed(77)

# %%
# Example 3: using a previously saved generator state
# ---------------------------------------------------

# use the random generator
ot.Normal().getSample(100)
# save the generator state
state = ot.RandomGenerator.GetState()
ot.Normal().getRealization()

# %%
# load the generator state
ot.RandomGenerator.SetState(state)
ot.Normal().getRealization()



"""
Gaussian Process Regression: nugget effect and noise
====================================================
"""

# %%
# In this example, we show how to estimate a Gaussian Process Regression surrogate model
# (refer to :ref:`gaussian_process_regression`)
# from a noisy data set of the model. We consider the following cases:
#
# - Case 1: No noise is considered, assuming that the output values are exact,
# - Case 2: We consider a homoscedastic noise on the output values,
# - Case 3: We consider a heteroscedastic noise on the output values,
# - Case 4: We consider a homoscedastic noise modeled as a nugget factor.
#
# Both Case 2 and Case 4 model a homoscedastic noise on the output values of the model. But the two models
# are quite different:
#
# - the :meth:`~openturns.GaussianProcessFitter.setNoise` method of the :class:`~openturns.GaussianProcessFitter`
#   takes noise into account only in
#   the expression of the model likelihood, in the evaluation of the covariance parameters,
# - the :meth:`~openturns.CovarianceModel.setNuggetFactor` method of the :class:`~openturns.CovarianceModel`
#   modifies the covariance model of
#   the Gaussian process and keeps it modified once the parameters have been estimated. As a result, the
#   trajectories of the Gaussian process become less smooth than with the first model.
#
# Modeling a noise on the output values as a nugget effect is interesting to
# simulate some sensors which have a finite precision and generates some output values perturbated by
# a White Noise process.
#
# On the contrary, taking noise into account with the :meth:`~openturns.GaussianProcessFitter.setNoise` method
# impacts the estimation of the covariance parameters
# but not the regularity of the process.

# %%
# Introduction
# ------------
#
# We consider the model :math:`\model: [0,12] \rightarrow \Rset` defined by:
#
# .. math::
#
#      y = \model (x) = x \sin(x)
#
#
# We want to create a surrogate of this model using Gaussian Process Regression, from the data set defined by:
#
# .. math::
#
#      y_k = x_k \sin(x_k), \quad 1 \leq k \leq \sampleSize
#

# %%
import openturns as ot
import openturns.experimental as otexp
import openturns.viewer as otv
import math as m

# %%
# First let us introduce some useful function.
# In order to observe the function and the location of the points in the input design of experiments, we define `plot_1d_data`.

# sphinx_gallery_thumbnail_number = 10


# %%
def plot_1d_data(x_data, y_data, type="Curve", legend=None, color=None, linestyle=None):
    """Plot the data (x_data,y_data) as a Cloud/Curve"""
    if type == "Curve":
        graphF = ot.Curve(x_data, y_data)
    else:
        graphF = ot.Cloud(x_data, y_data)
    if legend is not None:
        graphF.setLegend(legend)
    if color is not None:
        graphF.setColor(color)
    if linestyle is not None:
        graphF.setLineStyle(linestyle)
    return graphF


# %%
# We define the model to be approximated.
g = ot.SymbolicFunction(["x"], ["x * sin(x)"])

# %%
# We define a train data set.
xmin = 0.0
xmax = 12.0
n_train = 10
step = (xmax - 1 - xmin) / (n_train - 1.0)
x_train = ot.RegularGrid(xmin + 0.2, step, n_train).getVertices()
y_train = g(x_train)
n_train = x_train.getSize()

# %%
# In order to compare the function and its metamodel, we use a test (i.e. validation) design of experiments made
# of a regular grid of 100 points from 0 to 12.
# Then we convert this grid into a :class:`~openturns.Sample` and we compute the outputs of the function on this
# sample.

# %%
n_test = 500
step = (xmax - xmin) / (n_test - 1)
myRegularGrid = ot.RegularGrid(xmin, step, n_test)
x_test = myRegularGrid.getVertices()
y_test = g(x_test)

# %%
# We plot the model.
graph = ot.Graph(r"Model $x \rightarrow x\sin(x)$", "", "")
graph.add(
    plot_1d_data(x_test, y_test, legend="Model", color="black", linestyle="dashed")
)
graph.setAxes(True)
graph.setXTitle("X")
graph.setYTitle("Y")
graph.setLegendPosition("lower left")
view = otv.View(graph)

# %%
# We use the :class:`~openturns.ConstantBasisFactory` class to define the trend and the :class:`~openturns.MaternModel` class to define the covariance model.
# This Matérn model is based on the regularity parameter :math:`\nu=3/2`.

# %%
ot.ResourceMap.SetAsScalar("CovarianceModel-DefaultNuggetFactor", 0.0)
dimension = 1
basis = ot.ConstantBasisFactory(dimension).build()

# %%
# Case 1: Ignore any noise
# ~~~~~~~~~~~~~~~~~~~~~~~~
# We introduce no noise here: it means that we assume that the output values are exact.
# We plot the surrogate model: we see that it interpolates the data set, as expected.
covarianceModel = ot.MaternModel([1.0] * dimension, 1.5)
fitter_algo = ot.GaussianProcessFitter(x_train, y_train, covarianceModel, basis)
fitter_algo.run()
fitter_result_noNoise = fitter_algo.getResult()

# %%
# We get the estimated parameters of the covariance model.
estimated_cov_noNoise = fitter_result_noNoise.getCovarianceModel()
print("Case 1: ignore any noise")
print(
    "Estimated covariance model with the homoscedastic noise = ", estimated_cov_noNoise
)

# %%
# We create the GPR metamodel.
gpr_algo_noNoise = ot.GaussianProcessRegression(fitter_result_noNoise)
gpr_algo_noNoise.run()
gpr_result_noNoise = gpr_algo_noNoise.getResult()
gprMM_noNoise = gpr_result_noNoise.getMetaModel()

graph = ot.Graph("Model with exact data", "x", "y")
graph.setLegendPosition("lower left")
graph.add(
    plot_1d_data(x_test, y_test, legend="Model", color="black", linestyle="dashed")
)
graph.add(
    plot_1d_data(x_train, y_train, type="Cloud", legend="Exact Data", color="red")
)
graph.add(plot_1d_data(x_test, gprMM_noNoise(x_test), legend="GPR", color="green"))
view = otv.View(graph)

# %%
# We can draw some trajectories of the resulting conditioned Gaussian process using the class
# :class:`~openturns.experimental.ConditionedGaussianProcess` built from a :class:`~openturns.GaussianProcessRegressionResult`.
# We create the conditioned Gaussian process. Then we generate some trajectories.
process_noNoise = otexp.ConditionedGaussianProcess(gpr_result_noNoise, myRegularGrid)
traj_noNoise = process_noNoise.getSample(10)
graph_traj_noNoise = ot.Graph(r"Model with exact data", "x", "y", True, "lower left")
graph_traj_noNoise.add(traj_noNoise.drawMarginal())
graph_traj_noNoise.setLegends(["conditioned trajectories"] + [""] * 9)
graph_traj_noNoise.setColors(["grey"] * 10)
graph_traj_noNoise.add(
    plot_1d_data(x_test, y_test, legend="Model", color="black", linestyle="dashed")
)
graph_traj_noNoise.add(
    plot_1d_data(x_test, gprMM_noNoise(x_test), legend="GPR", color="green")
)
graph_traj_noNoise.add(
    plot_1d_data(x_train, y_train, type="Cloud", legend="Exact data", color="red")
)
view = otv.View(graph_traj_noNoise)

# %%
# Case 2: Introduce a homoscedastic noise
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# We want to model a noise on each output value of the model. This noise is the same
# for output :math:`y_i` value and is modeled as a :class:`~openturns.Normal` distribution
# which is centered with a variance denoted by :math:`\sigma^2 = 0.5`.
variance_point_homosc = [0.25] * n_train

# %%
# In order to illustrate some noisy output values, we add a random noise to each
# output, following the noise distribution.
noise_homosk = ot.Normal(0, m.sqrt(variance_point_homosc[0])).getSample(n_train)
y_train_homosc = y_train + noise_homosk

# %%
# We plot the model and the noisy train data
# which is not on the graph of the model.
graph = ot.Graph(r"Model with homoscedastic noisy data", "x", "y", True, "lower left")
graph.add(
    plot_1d_data(x_test, y_test, legend="Model", color="black", linestyle="dashed")
)
graph.add(
    plot_1d_data(
        x_train, y_train_homosc, type="Cloud", legend="Noisy data", color="red"
    )
)
view = otv.View(graph)

# %%
# We define the Gaussian Process Fitter on the noisy data.
covarianceModel = ot.MaternModel([1.0] * dimension, 1.5)
fitter_algo = ot.GaussianProcessFitter(x_train, y_train_homosc, covarianceModel, basis)

# %%
# If we ignore the noise, we get the following surrogate model.
fitter_algo.run()
fitter_result = fitter_algo.getResult()

# %%
# We get the estimated parameters of the covariance model.
estimated_cov = fitter_result.getCovarianceModel()
print("Case 2: homoscedastic noise")
print("Estimated covariance model with the homoscedastic noise = ", estimated_cov)

# %%
# We create the GPR metamodel.
gpr_algo_1 = ot.GaussianProcessRegression(fitter_result)
gpr_algo_1.run()
gpr_result_1 = gpr_algo_1.getResult()
gprMM_1 = gpr_result_1.getMetaModel()


# %%
# Now we do not ignore the noise on the output values any more and we add it to the algorithm.
covarianceModel = ot.MaternModel([1.0] * dimension, 1.5)
fitter_algo = ot.GaussianProcessFitter(x_train, y_train_homosc, covarianceModel, basis)
fitter_algo.setNoise(variance_point_homosc)
fitter_algo.run()
fitter_result = fitter_algo.getResult()
gpr_algo_homosc = ot.GaussianProcessRegression(fitter_result)
gpr_algo_homosc.run()
gpr_result_homosc = gpr_algo_homosc.getResult()
gprMM_homosc = gpr_result_homosc.getMetaModel()

# %%
# We plot now the resulting surrogate model. We can see that the GPR metamodel
# does not interpolate the train set any more due to the introduction of the
# noise.
graph = ot.Graph(r"Model with homoscedastic noisy data", "x", "y", True, "lower left")
graph.add(
    plot_1d_data(x_test, y_test, legend="Model", color="black", linestyle="dashed")
)
graph.add(
    plot_1d_data(
        x_train, y_train_homosc, type="Cloud", legend="Noisy data", color="red"
    )
)
graph.add(
    plot_1d_data(
        x_test, gprMM_homosc(x_test), legend="GPR modeling noise", color="green"
    )
)
graph.add(
    plot_1d_data(x_test, gprMM_1(x_test), legend="GPR ignoring noise", color="blue")
)
view = otv.View(graph)

# %%
# We can draw some trajectories of the resulting conditioned Gaussian process still using the class
# :class:`~openturns.experimental.ConditionedGaussianProcess` as previoulsy.
process_homosc = otexp.ConditionedGaussianProcess(gpr_result_homosc, myRegularGrid)
traj_homosc = process_homosc.getSample(10)
graph_traj_homosc = ot.Graph(
    r"Model with homoscedastic noisy data", "x", "y", True, "lower left"
)
graph_traj_homosc.add(traj_homosc.drawMarginal())
graph_traj_homosc.setLegends(["conditioned trajectories"] + [""] * 9)
graph_traj_homosc.setColors(["grey"] * 10)
graph_traj_homosc.add(
    plot_1d_data(x_test, y_test, legend="Model", color="black", linestyle="dashed")
)
graph_traj_homosc.add(
    plot_1d_data(
        x_test, gprMM_homosc(x_test), legend="GPR modeling noise", color="green"
    )
)
graph_traj_homosc.add(
    plot_1d_data(
        x_train, y_train_homosc, type="Cloud", legend="Noisy data", color="red"
    )
)
view = otv.View(graph_traj_homosc)

# %%
# Case 3: Introduce a heteroscedastic noise
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# We want to model a noise on each output value of the model. This noise depends on the
# output value :math:`y_i`  and is modeled as a :class:`~openturns.Normal` distribution
# which is centered with a variance denoted by :math:`\sigma_i^2`.
# We generate a list of :math:`\sampleSize` variances.
variance_point_heterosc = ot.Uniform(0.0, 1.5).getSample(n_train).asPoint()

# %%
# Once more, in order to illustrate some noisy output values, we add a random noise to each
# output, following the noise distribution.
noise_heterosk = ot.Sample(0, 1)
for i in range(n_train):
    noise_i = ot.Normal(0, m.sqrt(variance_point_heterosc[i])).getRealization()
    noise_heterosk.add(noise_i)
y_train_heterosc = y_train + noise_heterosk

# %%
# We plot the model and the noisy train data
# which is not on the graph of the model.
graph = ot.Graph(r"Model with heteroscedastic noisy data", "x", "y", True, "lower left")
graph.add(
    plot_1d_data(x_test, y_test, legend="Model", color="black", linestyle="dashed")
)
graph.add(
    plot_1d_data(
        x_train, y_train_heterosc, type="Cloud", legend="Noisy data", color="red"
    )
)
view = otv.View(graph)

# %%
# We define the Gaussian Process Fitter on the noisy data.
covarianceModel = ot.MaternModel([1.0] * dimension, 1.5)
fitter_algo = ot.GaussianProcessFitter(
    x_train, y_train_heterosc, covarianceModel, basis
)

# %%
# If we ignore the noise, we get the following surrogate model.
fitter_algo.run()
fitter_result = fitter_algo.getResult()

# %%
# We get the estimated parameters of the covariance model.
estimated_cov = fitter_result.getCovarianceModel()
print("Case 3: heteroscedastic noise")
print("Estimated covariance model with the heteroscedastic noise = ", estimated_cov)

# %%
# We create the GPR metamodel.
gpr_algo_2 = ot.GaussianProcessRegression(fitter_result)
gpr_algo_2.run()
gpr_result_2 = gpr_algo_2.getResult()
gprMM_2 = gpr_result_2.getMetaModel()

# %%
# Now we do not ignore the noise on output values and we add it to the algorithm.
fitter_algo = ot.GaussianProcessFitter(
    x_train, y_train_heterosc, covarianceModel, basis
)
fitter_algo.setNoise(variance_point_heterosc)
fitter_algo.run()
fitter_result = fitter_algo.getResult()
gpr_algo_heterosc = ot.GaussianProcessRegression(fitter_result)
gpr_algo_heterosc.run()
gpr_result_heterosc = gpr_algo_heterosc.getResult()
gprMM_heterosc = gpr_result_heterosc.getMetaModel()

# %%
# We plot now the resulting surrogate model. We can see that the Gaussian process
# regression does not interpolate the train set any more due to the introduction of the
# noise.
# We could compute the mean square error betwwen the model and the metamodels and show that the distribution of
# the errors is more tight when we take into account the noise modeling.
graph = ot.Graph(r"Model with heteroscedastic noisy data", "x", "y", True, "lower left")
graph.add(
    plot_1d_data(x_test, y_test, legend="Model", color="black", linestyle="dashed")
)
graph.add(
    plot_1d_data(
        x_train, y_train_heterosc, type="Cloud", legend="Noisy data", color="red"
    )
)
graph.add(
    plot_1d_data(
        x_test, gprMM_heterosc(x_test), legend="GPR modeling noise", color="green"
    )
)
graph.add(
    plot_1d_data(x_test, gprMM_2(x_test), legend="GPR ignoring noise", color="blue")
)
view = otv.View(graph)


# %%
# We can draw some trajectories of the resulting conditioned Gaussian process as previoulsy.
# We can see that the trajectories have kept the regularity of the initial Matern model.
process_heterosc = otexp.ConditionedGaussianProcess(gpr_result_heterosc, myRegularGrid)
traj_heterosc = process_heterosc.getSample(10)
graph_traj_heterosc = ot.Graph(
    r"Model with heteroscedastic noisy data", "x", "y", True, "lower left"
)
graph_traj_heterosc.add(traj_heterosc.drawMarginal())
graph_traj_heterosc.setLegends(["conditioned trajectories"] + [""] * 9)
graph_traj_heterosc.setColors(["grey"] * 10)
graph_traj_heterosc.add(
    plot_1d_data(x_test, y_test, legend="Model", color="black", linestyle="dashed")
)
graph_traj_heterosc.add(
    plot_1d_data(
        x_test, gprMM_heterosc(x_test), legend="GPR modeling noise", color="green"
    )
)
graph_traj_heterosc.add(
    plot_1d_data(
        x_train, y_train_heterosc, type="Cloud", legend="Noisy data", color="red"
    )
)
view = otv.View(graph_traj_heterosc)

# %%
# Case 4: Introduce a nugget effect
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# We can introduce the homoscedastic noise on the output values of the model as a nugget effect,
# using the :meth:`~openturns.CovarianceModel.setNuggetFactor` method of the
# :class:`~openturns.CovarianceModel`.
#
# We consider the Matern covariance model defined previously and we activate the estimation
# of the nugget factor.
covarianceModel = ot.MaternModel([1.0] * dimension, 1.5)
covarianceModel.activateNuggetFactor(True)
fitter_algo_nugget = ot.GaussianProcessFitter(
    x_train, y_train_homosc, covarianceModel, basis
)
fitter_algo_nugget.run()
fitter_result_nugget = fitter_algo_nugget.getResult()

# %%
# We get the estimated parameters of the covariance model and the nugget factor
# :math:`\varepsilon_{nugget}`.
estimated_cov_nugget = fitter_result_nugget.getCovarianceModel()
print("Case 4: nugget effect")
print("Estimated covariance model with the nugget factor = ", estimated_cov_nugget)
print("Estimated nugget factor = ", estimated_cov_nugget.getNuggetFactor())

# %%
gpr_algo_nuggetF = ot.GaussianProcessRegression(fitter_result_nugget)
gpr_algo_nuggetF.run()
gpr_result_nuggetF = gpr_algo_nuggetF.getResult()
gprMM_nuggetF = gpr_result_nuggetF.getMetaModel()

# %%
# We draw the graph to compare the nugget effect and the noise modeling.
graph = ot.Graph(r"Model with homoscedastic noisy data", "x", "y", True, "lower left")
graph.add(
    plot_1d_data(x_test, y_test, legend="Model", color="black", linestyle="dashed")
)
graph.add(
    plot_1d_data(
        x_train, y_train_homosc, type="Cloud", legend="Noisy data", color="red"
    )
)
graph.add(
    plot_1d_data(
        x_test,
        gprMM_nuggetF(x_test),
        legend="GPR modeling nugget factor",
        color="green",
    )
)
graph.add(
    plot_1d_data(
        x_test, gprMM_homosc(x_test), legend="GPR modeling noise", color="blue"
    )
)
view = otv.View(graph)


# %%
# We can draw some trajectories of the resulting conditioned Gaussian process as previoulsy.
# We can see that the trajectories are not as smooth as the initial Matern model any more.
process_nuggetF = otexp.ConditionedGaussianProcess(gpr_result_nuggetF, myRegularGrid)
traj_nuggetF = process_nuggetF.getSample(10)
graph_traj_nuggetF = ot.Graph(
    r"Model with homoscedastic noisy data", "x", "y", True, "lower left"
)
graph_traj_nuggetF.add(traj_nuggetF.drawMarginal())
graph_traj_nuggetF.setLegends(["conditioned trajectories"] + [""] * 9)
graph_traj_nuggetF.setColors(["grey"] * 10)
graph_traj_nuggetF.add(
    plot_1d_data(x_test, y_test, legend="Model", color="black", linestyle="dashed")
)
graph_traj_nuggetF.add(
    plot_1d_data(
        x_test,
        gprMM_nuggetF(x_test),
        legend="GPR modeling nugget factor",
        color="green",
    )
)
graph_traj_nuggetF.add(
    plot_1d_data(
        x_train, y_train_homosc, type="Cloud", legend="Noisy data", color="red"
    )
)
view = otv.View(graph_traj_nuggetF)

# %%
# To facilitate comparison, we draw the trajectories generated by the Gaussian process built with nugget factor
# or with noise.
grid = ot.GridLayout(1, 2)
grid.setGraph(0, 0, graph_traj_nuggetF)
grid.setGraph(0, 1, graph_traj_homosc)
view = otv.View(grid)

# %%
# Display all figures
otv.View.ShowAll()



"""
Customize your Metropolis-Hastings algorithm
============================================

This simple example shows how you can build your own variant
of the Metropolis-Hastings algorithm.
"""

# %%
# We want to sample from the distribution with support :math:`[-3, 3]^2`
# whose PDF :math:`f` is proportional to the Ackley function to the tenth power:
#
# .. math::
#     \forall \vect{x} \in [-3, 3]^2 \quad f(\vect{x}) \propto a(\vect{x})^{10},
#
# where :math:`a` is the Ackey function defined in :ref:`use-case-ackley` page.
# In the following we call it the "Ackley distribution".

import openturns as ot
import openturns.viewer as otv
from openturns.usecases import ackley_function
from numpy import exp, format_float_scientific

ot.RandomGenerator.SetSeed(100)

# %%
# Prepare the Metropolis-Hastings algorithm
# -----------------------------------------
# Define the Ackley distribution support and density (up to a constant factor).

am = ackley_function.AckleyModel()
ackley = am.model
power10 = ot.SymbolicFunction("x", "x^10")
ackley_pdf = ot.ComposedFunction(power10, ackley)
logarithm = ot.SymbolicFunction("x", "10 * log(x)")
ackley_logpdf = ot.ComposedFunction(logarithm, ackley)
lb = -3.0
ub = 3.0
support = ot.Interval([lb] * 2, [ub] * 2)

# %%
# Define the proposal distribution as a :class:`~openturns.Histogram`.
# Its ticks (on the X axis of the PDF of the histogram) will remain the same,
# but its frequencies (on the Y axis) will be updated
# during the course of the Metropolis-Hastings algorithm.

n_bins = 50
myticks = ot.RegularGrid(lb, (ub - lb) / n_bins, n_bins + 1).getValues()
frequencies = [1.0] * (myticks.getSize() - 1)
proposal = ot.Histogram(myticks, frequencies)

# %%
# The state of the Markov chain must be converted to an acceptable set
# of parameters for the :class:`~openturns.Histogram` distribution.
# This is the job of the *link function*,
# which we construct with the :class:`~openturns.OpenTURNSPythonFunction` class.
# It takes a state as input and outputs the parameters (ticks and frequencies)
# of the proposal distribution.
# In our case, the ticks will not depend on the inputs,
# but the frequencies will be outputs of the Ackley function.

parameter_dim = proposal.getParameter().getSize()
parameter_desc = proposal.getParameterDescription()


class ConditionalAckley(ot.OpenTURNSPythonFunction):
    """
    When executed, this function returns the parameters of a Histogram
    which approximates the conditional Ackley distribution obtained
    when one of the 2 coordinates is fixed.
    To compute the frequencies of the Histogram,
    this OpenTURNSPythonFunction computes the values of the Ackley function
    on a regular grid on the line parallel to
    either the (1, 0) vector (if the second coordinate is fixed)
    or the (0, 1) vector (if the first coordinate is fixed)
    containing the point passed as input.
    The regular grid covers the part of this line which is contained in
    the support of the Ackley distribution, implicitly defined as the
    smallest square that contains the cartesian product of the regular grid
    with itself. For example, if the regular grid covers the interval [-3, 3],
    then the support is the square [-3, 3] x [-3, 3].

    Parameters
    ----------
    marginal : int
        The marginal whose value is *not* fixed.
        If 0, then the line of the regular grid is parallel to the (1, 0) vector.
        If 1, then the line of the regular grid is parallel to the (0, 1) vector.
    ticks : RegularGrid
        Ticks of the Histogram distribution.

    """

    def __init__(self, marginal, ticks):
        super().__init__(2, parameter_dim)
        self.setInputDescription(["X0", "X1"])  # input: X0 and X1 coordinates
        self.setOutputDescription(parameter_desc)  # output: parameters of the Histogram
        self._marginal = marginal  # parameter which does not vary after initialization
        offset = (ticks[1] - ticks[0]) / 2
        self._marginal_inputs = ot.Sample.BuildFromPoint(ticks)[0:-1] + offset
        # _marginal_inputs contains the varying coordinate of the points in the regular grid
        self._size = self._marginal_inputs.getSize()
        self._ticks = ticks

    def _exec(self, X):
        """
        Execute the function on a point X = (X0, X1).

        Parameters
        ----------
        X : list of 2 floats
            Point through which the line containing the regular grid passes.

        Returns
        -------
        parameters : :class:`~openturns.Point`
            Parameters of the :class:`~openturns.Histogram`.

        """
        inputs = ot.Sample(self._size, X)  # sample of inputs for the Ackley function
        # All input points are initialized at the point X passed as argument.

        # Replace the varying coordinate with the values of the regular grid.
        inputs[:, self._marginal] = self._marginal_inputs

        # Compute the Ackley function on these inputs.
        outputs = exp(ackley_logpdf(inputs).asPoint())

        # The outputs are the unnormalized frequencies of the Histogram
        # proposal distribution, but the Histogram.setParameter() method
        # expects a full set of parameters.
        # The easiest way to provide it is to construct a new Histogram object
        # with the adequate frequencies and call its getParameter() method.
        return ot.Histogram(self._ticks, outputs).getParameter()


# %%
# The 2 components of the state of the Markov chain will be updated
# one after the other, not simultaneously.
# We define 2 :class:`~openturns.UserDefinedMetropolisHastings` algorithms
# encapsulated within a :class:`~openturns.Gibbs` algorithm,
# so we need 2 link functions, each corresponding to one of the marginals
# of the Ackley distribution.
#
# Note that thanks to the :class:`~openturns.OpenTURNSPythonFunction` class,
# we were able to only code one template to be used by two different functions
# instead of directly coding two functions with the :class:`~openturns.PythonFunction` class.

link_function_0 = ot.Function(ConditionalAckley(0, myticks))
link_function_1 = ot.Function(ConditionalAckley(1, myticks))

# %%
# Let us illustrate the first of these functions.
# We can start by evaluating it at :math:`(0.5, 1.5)`.
# Let :math:`(X_0, X_1)` be a bidimensional random variable
# following the Ackley distribution.
# The output is the set of parameters for a :class:`~openturns.Histogram`
# distribution that approximates the distribution of :math:`X_0`
# conditional on :math:`X_1 = 1.5`.
# Let us update the Histogram we created before with this set of parameters.

par = link_function_0([0.5, 1.5])
proposal.setParameter(par)
print(par)

# %%
# The `link_function_0` function computes histogram frequencies proportional
# to the values of the unnormalized Ackley distribution PDF
# along the :math:`X_1 = 1.5` line.

title = "Ackley PDF (up to a constant factor) and $X_1 = 1.5$ cross-section"
graph = ot.Graph(title, "$X_0$", "$X_1$")
line = ot.Curve([[-3, 1.5], [3, 1.5]], "black", "dashed", 2)
graph.add(line)
contour = ackley_pdf.draw([lb] * 2, [ub] * 2).getDrawable(0).getImplementation()
contour.setLabels(
    [format_float_scientific(float(v), precision=1) for v in contour.getLabels()]
)
contour.setColorMapNorm("log")
graph.add(contour)
view = otv.View(graph)

# %%
# Let us compare the histogram to the unnormalized PDF
# (we still need to rescale it to make it appear properly on the graph).

scaling = ot.SymbolicFunction("x", "1.3e-10 * x")
scaled_ackley_pdf = ot.ComposedFunction(scaling, ackley_pdf)
graph = proposal.drawPDF()
graph.setXTitle("")
graph.add(scaled_ackley_pdf.draw(0, 0, [0.0, 1.5], -3.0, 3.0, 100))
graph.setTitle("Conditional distribution of $X_0$ given $X_1 = 1.5$")
_ = otv.View(graph)

# %%
# Let us now do the same think with `link_function_1`.
# This time, the relevant cross-section is along the line :math:`X_1 = 0.5`.

par = link_function_1([0.5, 1.5])
proposal.setParameter(par)
print(par)

title = "Ackley PDF (up to a constant factor) and $X_0 = 0.5$ cross-section"
graph = ot.Graph(title, "$X_0$", "$X_1$")
line = ot.Curve([[0.5, -3], [0.5, 3]], "black", "dashed", 2)
graph.add(line)
graph.setLegendPosition("upper right")
contour = ackley_pdf.draw([lb] * 2, [ub] * 2).getDrawable(0).getImplementation()
contour.setLabels(
    [format_float_scientific(float(v), precision=1) for v in contour.getLabels()]
)
contour.setColorMapNorm("log")
graph.add(contour)
view = otv.View(graph)

scaling = ot.SymbolicFunction("x", "3.1e-10 * x")
scaled_ackley_pdf = ot.ComposedFunction(scaling, ackley_pdf)
graph = proposal.drawPDF()
graph.setXTitle("")
graph.add(scaled_ackley_pdf.draw(1, 0, [0.5, 0.0], -3.0, 3.0, 100))
graph.setTitle("Conditional distribution of $X_1$ given $X_0 = 0.5$")
_ = otv.View(graph)


# %%
# We now choose the initial state of the Markov chain.

initialState = [0.1] * 2

# %%
# Sample from the Ackley distribution
# -----------------------------------
# We can finally define the two Metropolis-Hastings algorithms
# and the Gibbs algorithm which encapsulates them.

gmh_0 = ot.UserDefinedMetropolisHastings(
    ackley_logpdf, support, initialState, proposal, link_function_0, [0]
)
gmh_1 = ot.UserDefinedMetropolisHastings(
    ackley_logpdf, support, initialState, proposal, link_function_1, [1]
)
gibbs = ot.Gibbs([gmh_0, gmh_1])

sample = gibbs.getSample(100)

# %%
# Print the acceptance rates of the two Metropolis-Hastings samplers
mhlist = gibbs.getMetropolisHastingsCollection()
rate_gmh_0 = mhlist[0].getAcceptanceRate()
rate_gmh_1 = mhlist[1].getAcceptanceRate()

print("Acceptance rates: {} and {}".format(rate_gmh_0, rate_gmh_1))

# %%
# Draw the contour plot of the Ackley distribution PDF (up to a multiplicative constant)
# and the MCMC sample: we can see the sample is accurate.

title = "Ackley PDF (up to a constant factor) and Metropolis-Hastings sample"
graph = ot.Graph(title, "$X_0$", "$X_1$")
contour = ackley_pdf.draw([lb] * 2, [ub] * 2).getDrawable(0).getImplementation()
contour.setLabels(
    [format_float_scientific(float(v), precision=1) for v in contour.getLabels()]
)
print(contour.getLabels())
contour.setColorMapNorm("log")
graph.add(contour)
graph.add(ot.Cloud(sample, "black", "plus"))
view = otv.View(graph)

# %%
# Display all figures
otv.View.ShowAll()



import openturns as ot
import openturns.viewer as otv


size = 100
inputDimension = 6
inputSample = ot.Normal(inputDimension).getSample(size)
inputVar = ["X" + str(i) for i in range(inputDimension)]
inputSample.setDescription(inputVar)
expression = ""
for i in range(inputDimension):
    if i > 0:
        expression += "+"
    expression += "cos(" + str(i + 1) + "*" + inputVar[i] + ")"
model = ot.SymbolicFunction(inputVar, [expression])
outputSample = model(inputSample)

cobweb = ot.VisualTest.DrawParallelCoordinates(
    inputSample, outputSample, 2.5, 3.0, "red", False
)

cobweb.setLegendCorner([1.0, 0.0])
cobweb.setLegendPosition("lower right")
otv.View(cobweb, figure_kw={"figsize": (9, 5)})
